#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi


class GerritGroupMembers(object):
    def __init__(self, group_id, gerrit):
        self.group_id = group_id
        self.gerrit = gerrit
        self.endpoint = f"/groups/{self.group_id}/members"

    def list(self):
        """
        Lists the direct members of a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :return:
        """
        result = self.gerrit.get(self.endpoint)

        accounts = []
        for item in result:
            username = item.get("username")
            accounts.append(self.gerrit.accounts.get(username))

        return accounts

    def get(self, username):
        """
        Retrieves a group member.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :param username: account username
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{username}")
        if result:
            username = result.get("username")
            return self.gerrit.accounts.get(username)

    def add(self, username):
        """
        Adds a user as member to a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :param username: account username
        :return:
        """
        result = self.gerrit.put(self.endpoint + f"/{username}")
        if result:
            username = result.get("username")
            return self.gerrit.accounts.get(username)

    def remove(self, username):
        """
        Removes a user from a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :param username: account username
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{username}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel
from gerrit.groups.members import GerritGroupMembers
from gerrit.groups.subgroups import GerritGroupSubGroups


class GerritGroup(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "name"
        self.endpoint = f"/groups/{self.id}"

    def get_name(self):
        """
        Retrieves the name of a group.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/name")

    def set_name(self, input_):
        """
        Renames a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        .. code-block:: python

            input_ = {
                "name": "My Project Committers"
            }

            group = client.groups.get('0017af503a22f7b3fa6ce2cd3b551734d90701b4')
            result = group.set_name(input_)

        :param input_:
        :return:
        """
        return self.gerrit.put(
            self.endpoint + "/name", json=input_, headers=self.gerrit.default_headers)

    def get_description(self):
        """
        Retrieves the description of a group.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/description")

    def set_description(self, input_):
        """
        Sets the description of a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        .. code-block:: python

            input_ = {
                "description": "The committers of MyProject."
            }
            group = client.groups.get('0017af503a22f7b3fa6ce2cd3b551734d90701b4')
            result = group.set_description(input_)

        :param input_:
        :return:
        """
        return self.gerrit.put(
            self.endpoint + "/description", json=input_, headers=self.gerrit.default_headers)

    def delete_description(self):
        """
        Deletes the description of a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :return:
        """
        self.gerrit.delete(self.endpoint + "/description")

    def get_options(self):
        """
        Retrieves the options of a group.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/options")

    def set_options(self, input_):
        """
        Sets the options of a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        .. code-block:: python

            input_ = {
                "visible_to_all": True
            }
            group = client.groups.get('0017af503a22f7b3fa6ce2cd3b551734d90701b4')
            result = group.set_options(input_)


        :param input_: the GroupOptionsInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-groups.html
          #group-options-input
        :return:
        """
        return self.gerrit.put(
            self.endpoint + "/options", json=input_, headers=self.gerrit.default_headers)

    def get_owner(self):
        """
        Retrieves the owner group of a Gerrit internal group.

        :return: As response a GroupInfo entity is returned that describes the owner group.
        """
        return self.gerrit.get(self.endpoint + "/owner")

    def set_owner(self, input_):
        """
        Sets the owner group of a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        .. code-block:: python

            input_ = {
                "owner": "6a1e70e1a88782771a91808c8af9bbb7a9871389"
            }
            group = client.groups.get('0017af503a22f7b3fa6ce2cd3b551734d90701b4')
            result = group.set_owner(input_)

        :param input_: As response a GroupInfo entity is returned that describes the new owner
        group.
        :return:
        """
        return self.gerrit.put(
            self.endpoint + "/owner", json=input_, headers=self.gerrit.default_headers)

    def get_audit_log(self):
        """
        Gets the audit log of a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/log.audit")

    def index(self):
        """
        Adds or updates the internal group in the secondary index.

        :return:
        """
        self.gerrit.post(self.endpoint + "/index")

    @property
    def members(self):
        return GerritGroupMembers(group_id=self.id, gerrit=self.gerrit)

    @property
    def subgroup(self):
        return GerritGroupSubGroups(group_id=self.id, gerrit=self.gerrit)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
import json
import netrc
from gerrit.utils.requester import Requester
from gerrit.config.config import GerritConfig
from gerrit.projects.projects import GerritProjects
from gerrit.accounts.accounts import GerritAccounts
from gerrit.groups.groups import GerritGroups
from gerrit.plugins.plugins import GerritPlugins
from gerrit.changes.changes import GerritChanges


class GerritClient(object):
    """
    Python wrapper for the Gerrit V3.x REST API.

    """

    default_headers = {"Content-Type": "application/json; charset=UTF-8"}

    def __init__(
        self,
        base_url,
        username,
        password=None,
        use_netrc=False,
        ssl_verify=True,
        cert=None,
        timeout=60,
        max_retries=None,
        auth_suffix="/a"
    ):
        if not password and not use_netrc:
            raise ValueError(
                "One of 'password' or 'use_netrc' parameters should be set!"
            )

        self._base_url = self.strip_trailing_slash(base_url)

        if not password and use_netrc:
            password = self.get_password_from_netrc_file()

        self.requester = Requester(
            base_url=base_url,
            username=username,
            password=password,
            ssl_verify=ssl_verify,
            cert=cert,
            timeout=timeout,
            max_retries=max_retries,
        )
        self.auth_suffix = auth_suffix

    def get_password_from_netrc_file(self):
        """
        Providing the password form .netrc file for getting Host name.
        :return: The related password from .netrc file as a string.
        """

        netrc_client = netrc.netrc()
        auth_tokens = netrc_client.authenticators(self._base_url)
        if not auth_tokens:
            raise ValueError(f"The '{self._base_url}' host name is not found in netrc file.")
        return auth_tokens[2]

    @classmethod
    def strip_trailing_slash(cls, url):
        """
        remove url's trailing slash
        :param url: url
        :return:
        """
        while url.endswith("/"):
            url = url[:-1]
        return url

    def get_endpoint_url(self, endpoint):
        """
        Return the complete url including host and port for a given endpoint.
        :param endpoint: service endpoint as str
        :return: complete url (including host and port) as str
        """
        return f"{self._base_url}{self.auth_suffix}{endpoint}"

    @staticmethod
    def decode_response(response):
        """Strip off Gerrit's magic prefix and decode a response.
        :returns:
            Decoded JSON content as a dict, or raw text if content could not be
            decoded as JSON.
        :raises:
            requests.HTTPError if the response contains an HTTP error status code.
        """
        magic_json_prefix = ")]}'\n"
        content_type = response.headers.get("content-type", "")

        content = response.content.strip()
        if response.encoding:
            content = content.decode(response.encoding)
        if not content:
            return content
        if content_type.split(";")[0] != "application/json":
            return content
        if content.startswith(magic_json_prefix):
            index = len(magic_json_prefix)
            content = content[index:]
        try:
            return json.loads(content)
        except ValueError:
            raise ValueError(f"Invalid json content: {content}")

    @property
    def config(self):
        """
        Config related REST APIs

        :return:
        """
        return GerritConfig(gerrit=self)

    @property
    def projects(self):
        """
        Project related REST APIs
        :return:
        """
        return GerritProjects(gerrit=self)

    @property
    def changes(self):
        """
        Change related REST APIs

        :return:
        """
        return GerritChanges(gerrit=self)

    @property
    def accounts(self):
        """
        Account related REST APIs

        :return:
        """
        return GerritAccounts(gerrit=self)

    @property
    def groups(self):
        """
        Group related REST APIs

        :return:
        """
        return GerritGroups(gerrit=self)

    @property
    def plugins(self):
        """
        Plugin related REST APIs

        :return:
        """
        return GerritPlugins(gerrit=self)

    @property
    def version(self):
        """
        get the version of the Gerrit server.

        :return:
        """
        return self.config.get_version()

    @property
    def server(self):
        """
        get the information about the Gerrit server configuration.

        :return:
        """
        return self.config.get_server_info()

    def get(self, endpoint, **kwargs):
        """
        Send HTTP GET to the endpoint.

        :param endpoint: The endpoint to send to.
        :return:
        """
        response = self.requester.get(self.get_endpoint_url(endpoint), **kwargs)
        result = self.decode_response(response)
        return result

    def post(self, endpoint, **kwargs):
        """
        Send HTTP POST to the endpoint.

        :param endpoint: The endpoint to send to.
        :return:
        """
        response = self.requester.post(self.get_endpoint_url(endpoint), **kwargs)
        result = self.decode_response(response)
        return result

    def put(self, endpoint, **kwargs):
        """
        Send HTTP PUT to the endpoint.

        :param endpoint: The endpoint to send to.
        :return:
        """
        response = self.requester.put(self.get_endpoint_url(endpoint), **kwargs)
        result = self.decode_response(response)
        return result

    def delete(self, endpoint):
        """
        Send HTTP DELETE to the endpoint.

        :param endpoint: The endpoint to send to.
        :return:
        """
        self.requester.delete(self.get_endpoint_url(endpoint))



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from copy import deepcopy


class ResultSet(list):
    """A list like object that holds results from a Gerrit API query."""


class BaseModel(object):
    def __init__(self, json=None, **kwargs):
        self.entity_name = "id"
        self.content = None

        # Parse a JSON object into a model instance.
        if json is not None:
            self.content = deepcopy(json)
            json.update(kwargs)

            for key, value in json.items():
                try:
                    if key[0] == '_':
                        key = key[1:]
                    setattr(self, key, value)
                except AttributeError:
                    pass

    def to_dict(self):
        return self.content

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__.get(key)
        else:
            raise AttributeError(key)

    @classmethod
    def parse_list(cls, data, **kwargs):
        """Parse a list of JSON objects into a result set of model instances."""
        results = ResultSet()
        data = data or []
        for obj in data:
            if obj:
                results.append(cls(json=obj, **kwargs))
        return results

    @classmethod
    def parse_dict(cls, data, **kwargs):
        """Parse a dict of JSON objects into a result set of model instances."""
        results = ResultSet()
        data = data or {}
        for obj in data.keys():
            if obj:
                results.append(cls(json=data[obj], **kwargs))
        return results

    def __repr__(self):
        key = self.entity_name
        value = getattr(self, key)
        return f"{self.__class__.__name__}({key}={value})"



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi

from gerrit.utils.models import BaseModel


class GerritProjectWebHook(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "name"
        self.endpoint = f"/config/server/webhooks~projects/{self.project}/remotes"

    def delete(self):
        """
        Delete a webhook for a project.

        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{self.name}")


class GerritProjectWebHooks(object):
    def __init__(self, project, gerrit):
        self.project = project
        self.gerrit = gerrit
        self.endpoint = f"/config/server/webhooks~projects/{self.project}/remotes"

    def list(self):
        """
        List existing webhooks for a project.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/")

        webhooks = []
        for key, value in result.items():
            webhook = value
            webhook.update({"name": key})
            webhooks.append(webhook)

        return GerritProjectWebHook.parse_list(
            webhooks, project=self.project, gerrit=self.gerrit
        )

    def create(self, name, input_):
        """
        Create or update a webhook for a project.

        .. code-block:: python

            input_ = {
                "url": "https://foo.org/gerrit-events",
                "maxTries": "3",
                "sslVerify": "true"
            }

            project = client.projects.get('myproject')
            new_webhook = project.webhooks.create('test', input_)

        :param name: the webhook name
        :param input_: the RemoteInfo entity
        :return:
        """
        result = self.gerrit.put(
            self.endpoint + f"/{name}", json=input_, headers=self.gerrit.default_headers)
        return GerritProjectWebHook(json=result, project=self.project, gerrit=self.gerrit)

    def get(self, name):
        """
        Get information about one webhook.

        :param name: the webhook name
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{name}")
        result.update({"name": name})
        return GerritProjectWebHook(json=result, project=self.project, gerrit=self.gerrit)

    def delete(self, name):
        """
        Delete a webhook for a project.

        :param name: the webhook name
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{name}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel


class Task(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = "/config/server/tasks"

    def delete(self):
        """
        Kills a task from the background work queue that the Gerrit daemon is currently performing,
        or will perform in the near future.

        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{self.id}")


class Tasks(object):
    def __init__(self, gerrit):
        self.gerrit = gerrit
        self.endpoint = "/config/server/tasks"

    def list(self):
        """
        Lists the tasks from the background work queues that the Gerrit daemon is currently
        performing, or will perform in the near future.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        return Task.parse_list(result, gerrit=self.gerrit)

    def get(self, id_):
        """
        Retrieves a task from the background work queue that the Gerrit daemon is currently
        performing, or will perform in the near future.

        :param id_: task id
        :return:
        """

        result = self.gerrit.get(self.endpoint + f"/{id_}")
        return Task(json=result, gerrit=self.gerrit)

    def delete(self, id_):
        """
        Kills a task from the background work queue that the Gerrit daemon is currently performing,
        or will perform in the near future.

        :param id_: task id
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{id_}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.common import params_creator
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from gerrit.utils.models import BaseModel


class GerritProjectTag(BaseModel):
    tag_prefix = "refs/tags/"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "ref"
        self.endpoint = f"/projects/{self.project}/tags"

    @property
    def name(self):
        return self.ref.replace(self.tag_prefix, "")

    def delete(self):
        """
        Delete a tag.

        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{self.name}")


class GerritProjectTags(object):
    tag_prefix = "refs/tags/"

    def __init__(self, project, gerrit):
        self.project = project
        self.gerrit = gerrit
        self.endpoint = f"/projects/{self.project}/tags"

    def list(self, pattern_dispatcher=None, limit=None, skip=None):
        """
        List the tags of a project.

        :param pattern_dispatcher: Dict of pattern type with respective
               pattern value: {('match'|'regex') : value}
        :param limit: Limit the number of tags to be included in the results.
        :param skip: Skip the given number of tags from the beginning of the list.
        :return:
        """
        params = params_creator((("n", limit), ("s", skip)),
                                {"match": "m", "regex": "r"}, pattern_dispatcher)
        return self.gerrit.get(self.endpoint + "/", params=params)

    def get(self, name):
        """
        get a tag by ref

        :param name: the tag ref
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{quote_plus(name)}")
        return GerritProjectTag(json=result, project=self.project, gerrit=self.gerrit)

    def create(self, name, input_):
        """
        Creates a new tag on the project.

        .. code-block:: python

            input_ = {
                "message": "annotation",
                'revision': 'c83117624b5b5d8a7f86093824e2f9c1ed309d63'
            }

            project = client.projects.get('myproject')
            new_tag = project.tags.create('1.1.8', input_)

        :param name: the tag name
        :param input_: the TagInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#tag-input
        :return:
        """
        return self.gerrit.put(
            self.endpoint + f"/{name}", json=input_, headers=self.gerrit.default_headers)

    def delete(self, name):
        """
        Delete a tag.

        :param name: the tag ref
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{quote_plus(name)}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.changes.change import GerritChange


class GerritChanges(object):
    def __init__(self, gerrit):
        self.gerrit = gerrit
        self.endpoint = "/changes"

    def search(self, query, options=None, limit=None, skip=None):
        """
        Queries changes visible to the caller.

        .. code-block:: python

            query = "is:open+owner:self+is:mergeable"
            result = client.changes.search(query=query, options=["LABELS"])

        :param query: Queries as a list of string
        :param options: List of options to fetch additional data about changes
        :param limit: Int value that allows to limit the number of changes
                      to be included in the output results
        :param skip: Int value that allows to skip the given number of
                     changes from the beginning of the list
        :return:
        """
        params = {
            k: v
            for k, v in (("o", options), ("n", limit), ("S", skip))
            if v is not None
        }

        return self.gerrit.get(self.endpoint + f"/?q={query}", params=params)

    def get(self, id_, detailed=False, options=None):
        """
        Retrieves a change.

        :param id_: change id
        :param detailed: boolean value, if True then retrieve a change with
                         labels, detailed labels, detailed accounts,
                         reviewer updates, and messages.
        :param options: List of options to fetch additional data about a change
        :return:
        """

        endpoint = self.endpoint + f"/{id_}/"
        if detailed:
            endpoint += "detail"

        result = self.gerrit.get(endpoint, params={"o": options})
        return GerritChange(json=result, gerrit=self.gerrit)

    def create(self, input_):
        """
        create a change

        .. code-block:: python

            input_ = {
                "project": "myProject",
                "subject": "Let's support 100% Gerrit workflow direct in browser",
                "branch": "stable",
                "topic": "create-change-in-browser",
                "status": "NEW"
            }
            result = client.changes.create(input_)

        :param input_: the ChangeInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#change-input
        :return:
        """
        result = self.gerrit.post(self.endpoint + '/', json=input_, headers=self.gerrit.default_headers)
        return GerritChange(json=result, gerrit=self.gerrit)

    def delete(self, id_):
        """
        Deletes a change.

        :param id_: change id
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{id_}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.groups.group import GerritGroup
from gerrit.utils.common import params_creator
from packaging.version import parse


class GerritGroups(object):
    def __init__(self, gerrit):
        self.gerrit = gerrit
        self.endpoint = "/groups"

    def list(self, pattern_dispatcher=None, options=None, limit=None, skip=None):
        """
        Lists the groups accessible by the caller.

        :param pattern_dispatcher: Dict of pattern type with respective
                     pattern value: {('match'|'regex') : value}
        :param options: Additional fields can be obtained by adding o parameters,
                        each option requires more lookups and slows down the query response time to
                        the client so they are generally disabled by default. Optional fields are:
                          INCLUDES: include list of direct subgroups.
                          MEMBERS: include list of direct group members.
        :param limit: Int value that allows to limit the number of groups
                      to be included in the output results
        :param skip: Int value that allows to skip the given
                     number of groups from the beginning of the list
        :return:
        """
        params = params_creator((("o", options), ("n", limit), ("S", skip)),
                                {"match": "m", "regex": "r"}, pattern_dispatcher)

        return self.gerrit.get(self.endpoint, params=params)

    def search(self, query, options=None, limit=None, skip=None):
        """
        Query Groups

        :param query:
        :param options: Additional fields can be obtained by adding o parameters,
                        each option requires more lookups and slows down the query response time to
                        the client so they are generally disabled by default. Optional fields are:
                          INCLUDES: include list of direct subgroups.
                          MEMBERS: include list of direct group members.
        :param limit: Int value that allows to limit the number of groups
                      to be included in the output results
        :param skip: Int value that allows to skip the given
                     number of groups from the beginning of the list
        :return:
        """
        version = self.gerrit.version
        if parse(version) < parse("3.2.0"):
            endpoint = self.endpoint + f"/?query2={query}"
        else:
            endpoint = self.endpoint + f"/?query={query}"

        params = {
            k: v
            for k, v in (("o", options), ("limit", limit), ("start", skip))
            if v is not None
        }

        return self.gerrit.get(endpoint, params=params)

    def get(self, id_, detailed=False):
        """
        Retrieves a group.

        :param id_: group id, or group_id, or group name
        :param detailed:
        :return:
        """
        endpoint = self.endpoint + f"/{id_}/"
        if detailed:
            endpoint += "detail"
        return GerritGroup(json=self.gerrit.get(endpoint), gerrit=self.gerrit)

    def create(self, name, input_):
        """
        Creates a new Gerrit internal group.

        .. code-block:: python

            input_ = {
                "description": "contains all committers for MyProject2",
                "visible_to_all": 'true',
                "owner": "Administrators",
                "owner_id": "af01a8cb8cbd8ee7be072b98b1ee882867c0cf06"
            }
            new_group = client.groups.create('My-Project2-Committers', input_)

        :param name: group name
        :param input_: the GroupInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-groups.html#group-input
        :return:
        """
        result = self.gerrit.put(
            self.endpoint + f"/{name}", json=input_, headers=self.gerrit.default_headers)
        return GerritGroup(json=result, gerrit=self.gerrit)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi


class GerritGroupSubGroups(object):
    def __init__(self, group_id, gerrit):
        self.group_id = group_id
        self.gerrit = gerrit
        self.endpoint = f"/groups/{self.group_id}/groups/"

    def list(self):
        """
        Lists the direct subgroups of a group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        subgroups = []
        for item in result:
            group_id = item.get("id")
            subgroups.append(self.gerrit.groups.get(group_id))

        return subgroups

    def get(self, subgroup):
        """
        Retrieves a subgroup.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :param subgroup: subgroup id or name
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{subgroup}")
        if result:
            subgroup_id = result.get("id")
            return self.gerrit.groups.get(subgroup_id)

    def add(self, subgroup):
        """
        Adds an internal or external group as subgroup to a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :param subgroup: subgroup id or name
        :return:
        """
        result = self.gerrit.put(self.endpoint + f"/{subgroup}")
        if result:
            subgroup_id = result.get("id")
            return self.gerrit.groups.get(subgroup_id)

    def remove(self, subgroup):
        """
        Removes a subgroup from a Gerrit internal group.
        This endpoint is only allowed for Gerrit internal groups;
        attempting to call on a non-internal group will return 405 Method Not Allowed.

        :param subgroup: subgroup id or name
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{subgroup}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel


class Cache(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "name"
        self.endpoint = "/config/server/caches"

    def flush(self):
        """
        Flushes a cache.

        :return:
        """
        self.gerrit.post(self.endpoint + f"/{self.name}/flush")


class Caches(object):
    def __init__(self, gerrit):
        self.gerrit = gerrit
        self.endpoint = "/config/server/caches"

    def list(self):
        """
        Lists the caches of the server. Caches defined by plugins are included.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        caches = []
        for key, value in result.items():
            cache = value
            cache.update({"name": key})
            caches.append(cache)

        return Cache.parse_list(caches, gerrit=self.gerrit)

    def get(self, name):
        """
        Retrieves information about a cache.

        :param name: cache name
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{name}")
        return Cache(json=result, gerrit=self.gerrit)

    def flush(self, name):
        """
        Flushes a cache.

        :param name: cache name
        :return:
        """
        self.gerrit.post(self.endpoint + f"/{name}/flush")

    def operation(self, input_):
        """
        Cache Operations

        .. code-block:: python

            input_ = {
                "operation": "FLUSH_ALL"
            }
            gerrit.config.caches.operation(input_)

        :param input_: the CacheOperationInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-config.html
          #cache-operation-input
        :return:
        """
        self.gerrit.post(self.endpoint, json=input_, headers=self.gerrit.default_headers)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus
from gerrit.projects.project import GerritProject
from gerrit.utils.common import params_creator


class GerritProjects(object):
    def __init__(self, gerrit):
        self.gerrit = gerrit
        self.endpoint = "/projects"

    def list(
        self,
        is_all=False,
        limit=None,
        skip=None,
        pattern_dispatcher=None,
        project_type=None,
        description=False,
        branch=None,
        state=None,
    ):
        """
        Get list of all available projects accessible by the caller.

        :param is_all: boolean value, if True then all projects (including
                       hidden ones) will be added to the results.
                       May not be used together with the state option.
        :param limit: Int value that allows to limit the number of projects
                      to be included in the output results
        :param skip: Int value that allows to skip the given
                     number of projects from the beginning of the list
        :param pattern_dispatcher: Dict of pattern type with respective
                     pattern value: {('prefix'|'match'|'regex') : value}
        :param project_type: string value for type of projects to be fetched
                            ('code'|'permissions'|'all')
        :param description: boolean value, if True then description will be
                            added to the output result
        :param branch: Limit the results to the projects having the specified branch
                       and include the sha1 of the branch in the results.
        :param state: Get all projects with the given state. May not be used together with the all
        option.

        :return:
        """
        if is_all and state:
            raise ValueError("is_all can not be used together with the state option.")
        
        pattern_types = {"prefix": "p", "match": "m", "regex": "r"}
        tuples = (("n", limit),("S", skip),("type", project_type),("b", branch),("state", state))
        params = params_creator(tuples, pattern_types, pattern_dispatcher)
        params["all"] = int(is_all)
        params["d"] = int(description)

        return self.gerrit.get(self.endpoint + "/", params=params)

    def search(self, query, limit=None, skip=None):
        """
        Queries projects visible to the caller. The query string must be provided by the
        query parameter.
        The start and limit parameters can be used to skip/limit results.

        query parameter
          * name:'NAME' Matches projects that have exactly the name 'NAME'.
          * parent:'PARENT' Matches projects that have 'PARENT' as parent project.
          * inname:'NAME' Matches projects that a name part that starts with
           'NAME' (case insensitive).
          * description:'DESCRIPTION' Matches projects whose description contains 'DESCRIPTION',
          using a full-text search.
          * state:'STATE' Matches project’s state. Can be either 'active' or 'read-only'.

        :param query:
        :param limit: Int value that allows to limit the number of accounts
                      to be included in the output results
        :param skip: Int value that allows to skip the given
                     number of accounts from the beginning of the list
        :return:
        """
        params = {k: v for k, v in (("limit", limit), ("start", skip)) if v is not None}

        return self.gerrit.get(self.endpoint + f"/?query={query}", params=params)

    def get(self, name):
        """
        Retrieves a project.

        :param name: the name of the project
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{quote_plus(name)}")
        return GerritProject(json=result, gerrit=self.gerrit)

    def create(self, project_name, input_):
        """
        Creates a new project.

        .. code-block:: python

            input_ = {
                "description": "This is a demo project.",
                "submit_type": "INHERIT",
                "owners": [
                  "MyProject-Owners"
                ]
            }
            project = client.projects.create('MyProject', input_)

        :param project_name: the name of the project
        :param input_: the ProjectInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#project-input

        :return:
        """
        result = self.gerrit.put(
            self.endpoint + f"/{quote_plus(project_name)}",
            json=input_, headers=self.gerrit.default_headers)
        return GerritProject(json=result, gerrit=self.gerrit)

    def delete(self, project_name):
        """
        Delete the project, requires delete-project plugin

        :param project_name: project name
        :return:
        """
        self.gerrit.post(self.endpoint + f"/{quote_plus(project_name)}/delete-project~delete")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi

from gerrit.utils.models import BaseModel


class GerritProjectLabel(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "name"
        self.endpoint = f"/projects/{self.project}/labels/{self.name}"

    def set(self, input_):
        """
        Updates the definition of a label that is defined in this project.
        The calling user must have write access to the refs/meta/config branch of the project.
        Properties which are not set in the input entity are not modified.

        .. code-block:: python

            input_ = {
                "commit_message": "Ignore self approvals for Code-Review label",
                "ignore_self_approval": true
            }

            project = client.projects.get("MyProject")
            label = project.labels.get("foo")
            result = label.set(input_)

        :param input_: the LabelDefinitionInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #label-definition-input
        :return:
        """
        result = self.gerrit.put(self.endpoint, json=input_, headers=self.gerrit.default_headers)
        return self.gerrit.projects.get(self.project).labels.get(result.get("name"))

    def delete(self):
        """
        Deletes the definition of a label that is defined in this project.
        The calling user must have write access to the refs/meta/config branch of the project.

        :return:
        """
        self.gerrit.delete(self.endpoint)


class GerritProjectLabels(object):
    def __init__(self, project, gerrit):
        self.project = project
        self.gerrit = gerrit
        self.endpoint = f"/projects/{self.project}/labels"

    def list(self):
        """
        Lists the labels that are defined in this project.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/")
        return GerritProjectLabel.parse_list(result, gerrit=self.gerrit)

    def get(self, name):
        """
        Retrieves the definition of a label that is defined in this project.
        The calling user must have read access to the refs/meta/config branch of the project.

        :param name: label name
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{name}")
        return GerritProjectLabel(json=result, gerrit=self.gerrit)

    def create(self, name, input_):
        """
        Creates a new label definition in this project.
        The calling user must have write access to the refs/meta/config branch of the project.
        If a label with this name is already defined in this project, this label definition is
        updated (see Set Label).

        .. code-block:: python

            input_ = {
                "values": {
                    " 0": "No score",
                    "-1": "I would prefer this is not merged as is",
                    "-2": "This shall not be merged",
                    "+1": "Looks good to me, but someone else must approve",
                    "+2": "Looks good to me, approved"
                },
                "commit_message": "Create Foo Label"
            }
            new_label = project.labels.create('foo', input_)

        :param name: label name
        :param input_: the LabelDefinitionInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #label-definition-input
        :return:
        """
        result = self.gerrit.put(
            self.endpoint + f"/{name}", json=input_, headers=self.gerrit.default_headers)
        return GerritProjectLabel(json=result, gerrit=self.gerrit)

    def delete(self, name):
        """
        Deletes the definition of a label that is defined in this project.
        The calling user must have write access to the refs/meta/config branch of the project.

        :param name: label name
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{name}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from gerrit.utils.models import BaseModel
from gerrit.utils.common import params_creator


class GerritProjectBranch(BaseModel):
    branch_prefix = "refs/heads/"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "ref"
        self.endpoint = f"/projects/{self.project}/branches/{self.name}"

    @property
    def name(self):
        return self.ref.replace(self.branch_prefix, "")

    def get_file_content(self, file):
        """
        Gets the content of a file from the HEAD revision of a certain branch.
        The content is returned as base64 encoded string.

        :param file: the file path
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/files/{quote_plus(file)}/content")

    def is_mergeable(self, input_):
        """
        Gets whether the source is mergeable with the target branch.

        .. code-block:: python

            input_ = {
                'source': 'testbranch',
                'strategy': 'recursive'
            }
            result = stable.is_mergeable(input_)

        :param input_: the MergeInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#merge-input
        :return:
        """
        return self.gerrit.get(self.endpoint + "/mergeable", params=input_)

    def get_reflog(self):
        """
        Gets the reflog of a certain branch.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/reflog")

    def delete(self):
        """
        Delete a branch.

        :return:
        """
        self.gerrit.delete(self.endpoint)


class GerritProjectBranches(object):
    branch_prefix = "refs/heads/"

    def __init__(self, project, gerrit):
        self.project = project
        self.gerrit = gerrit
        self.endpoint = f"/projects/{self.project}/branches"

    def list(self, pattern_dispatcher=None, limit=None, skip=None):
        """
        List the branches of a project.

        :param pattern_dispatcher: Dict of pattern type with respective
               pattern value: {('match'|'regex') : value}
        :param limit: Limit the number of branches to be included in the results.
        :param skip: Skip the given number of branches from the beginning of the list.
        :return:
        """
        params = params_creator((("n", limit), ("s", skip)),
                                {"match": "m", "regex": "r"}, pattern_dispatcher)

        return self.gerrit.get(self.endpoint + "/", params=params)

    def get(self, name):
        """
        get a branch by ref

        :param name: branch ref name
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{quote_plus(name)}")
        return GerritProjectBranch(json=result, project=self.project, gerrit=self.gerrit)

    def create(self, name, input_):
        """
        Creates a new branch.

        .. code-block:: python

            input_ = {
                'revision': '76016386a0d8ecc7b6be212424978bb45959d668'
            }
            project = client.projects.get('myproject')
            new_branch = project.branches.create('stable', input_)


        :param name: the branch name
        :param input_: the BranchInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#branch-info
        :return:
        """
        return self.gerrit.put(
            self.endpoint + f"/{name}", json=input_, headers=self.gerrit.default_headers)

    def delete(self, name):
        """
        Delete a branch.

        :param name: branch ref name
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{quote_plus(name)}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel


class GerritAccountEmail(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "email"
        self.endpoint = f"/accounts/{self.username}/emails/{self.email}"

    def delete(self):
        """
        Deletes an email address of an account.

        :return:
        """
        self.gerrit.delete(self.endpoint)

    def set_preferred(self):
        """
        Sets an email address as preferred email address for an account.

        :return:
        """
        self.gerrit.put(self.endpoint + "/preferred")


class GerritAccountEmails(object):
    def __init__(self, username, gerrit):
        self.username = username
        self.gerrit = gerrit
        self.endpoint = f"/accounts/{self.username}/emails"

    def list(self):
        """
        Returns the email addresses that are configured for the specified user.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        return GerritAccountEmail.parse_list(result, username=self.username, gerrit=self.gerrit)

    def create(self, email):
        """
        Registers a new email address for the user.

        :return:
        """
        return self.gerrit.put(self.endpoint + f"/{email}")

    def get(self, email):
        """
        Retrieves an email address of a user.

        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{email}")
        return GerritAccountEmail(json=result, username=self.username, gerrit=self.gerrit)

    def set_preferred(self, email):
        """
        Sets an email address as preferred email address for an account.

        :param email: account email
        :return:
        """
        self.gerrit.put(self.endpoint + f"/{email}/preferred")

    def delete(self, email):
        """
        Deletes an email address of an account.

        :param email: account email
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{email}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel
from gerrit.utils.common import params_creator


class GerritPlugin(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = f"/plugins/{self.id}"

    def enable(self):
        """
        Enables a plugin on the Gerrit server.

        :return:
        """
        result = self.gerrit.post(self.endpoint + "/gerrit~enable")
        return self.gerrit.plugins.get(result.get("id"))

    def disable(self):
        """
        Disables a plugin on the Gerrit server.

        :return:
        """
        result = self.gerrit.post(self.endpoint + "/gerrit~disable")
        return self.gerrit.plugins.get(result.get("id"))

    def reload(self):
        """
        Reloads a plugin on the Gerrit server.

        :return:
        """
        result = self.gerrit.post(self.endpoint + "/gerrit~reload")
        return self.gerrit.plugins.get(result.get("id"))


class GerritPlugins(object):
    def __init__(self, gerrit):
        self.gerrit = gerrit
        self.endpoint = "/plugins"

    def list(self, is_all=False, limit=None, skip=None, pattern_dispatcher=None):
        """
        Lists the plugins installed on the Gerrit server.

        :param is_all: boolean value, if True then all plugins (including
                       hidden ones) will be added to the results
        :param limit: Int value that allows to limit the number of plugins
                      to be included in the output results
        :param skip: Int value that allows to skip the given
                     number of plugins from the beginning of the list
        :param pattern_dispatcher: Dict of pattern type with respective
                     pattern value: {('prefix'|'match'|'regex') : value}
        :return:
        """
        params = params_creator((("n", limit), ("S", skip)),
                                {"prefix": "p", "match": "m", "regex": "r"}, pattern_dispatcher)
        params["all"] = int(is_all)

        return self.gerrit.get(self.endpoint + "/", params=params)

    def get(self, id_):
        """

        :param id_: plugin id
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{id_}/gerrit~status")
        return GerritPlugin(json=result, gerrit=self.gerrit)

    def install(self, id_, input_):
        """
        Installs a new plugin on the Gerrit server.

        .. code-block:: python

            input_ = {
                "url": "file:///gerrit/plugins/delete-project/delete-project-2.8.jar"
            }

            plugin = client.plugins.install(input_)

        :param id_: plugin id
        :param input_: the PluginInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-plugins.html#plugin-input
        :return:
        """
        result = self.gerrit.put(
            self.endpoint + f"/{id_}.jar", json=input_, headers=self.gerrit.default_headers
        )
        return GerritPlugin(json=result, gerrit=self.gerrit)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from gerrit.utils.models import BaseModel
from gerrit.utils.exceptions import UnknownFile


class GerritChangeRevisionFile(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "path"
        self.endpoint = f"/changes/{self.change}/revisions/{self.revision}" \
                        f"/files/{quote_plus(self.path)}"

    def get_content(self):
        """
        Gets the content of a file from a certain revision.
        The content is returned as base64 encoded string.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/content")

    def download_content(self):
        """
        Downloads the content of a file from a certain revision, in a safe format that poses no
        risk for inadvertent execution of untrusted code.

        If the content type is defined as safe, the binary file content is returned verbatim.
        If the content type is not safe, the file is stored inside a ZIP file, containing a
        single entry with a random, unpredictable name having the same base and suffix as the
        true filename. The ZIP file is returned in verbatim binary form.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/download")

    def get_diff(self, intraline=False):
        """
        Gets the diff of a file from a certain revision.

        :param intraline: If the intraline parameter is specified, intraline differences are
        included in the diff.
        :return:
        """
        endpoint = self.endpoint + "/diff"
        if intraline:
            endpoint += endpoint + "?intraline"

        return self.gerrit.get(endpoint)

    def get_blame(self):
        """
        Gets the blame of a file from a certain revision.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/blame")

    def set_reviewed(self):
        """
        Marks a file of a revision as reviewed by the calling user.

        :return:
        """
        self.gerrit.put(self.endpoint + "/reviewed")

    def delete_reviewed(self):
        """
        Deletes the reviewed flag of the calling user from a file of a revision.

        :return:
        """
        self.gerrit.delete(self.endpoint + "/reviewed")


class GerritChangeRevisionFiles(object):
    def __init__(self, change, revision, gerrit):
        self.change = change
        self.revision = revision
        self.gerrit = gerrit
        self._data = []
        self.endpoint = f"/changes/{self.change}/revisions/{self.revision}/files"

    def poll(self):
        """

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        files = []
        for key, value in result.items():
            file = value
            file.update({"path": key})
            files.append(file)

        return files

    def iterkeys(self):
        """
        Iterate over the paths of all files

        :return:
        """
        if not self._data:
            self._data = self.poll()

        for file in self._data:
            yield file["path"]

    def keys(self):
        """
        Return a list of the file paths

        :return:
        """
        return list(self.iterkeys())

    def __len__(self):
        """

        :return:
        """
        return len(self.keys())

    def __contains__(self, ref):
        """

        :param ref:
        :return:
        """
        return ref in self.keys()

    def __iter__(self):
        """

        :return:
        """
        if not self._data:
            self._data = self.poll()

        for file in self._data:
            yield GerritChangeRevisionFile.parse(
                file, change=self.change, revision=self.revision, gerrit=self.gerrit
            )

    def __getitem__(self, path):
        """
        get a file by path

        :param path: file path
        :return:
        """
        if not self._data:
            self._data = self.poll()

        result = [file for file in self._data if file["path"] == path]
        if result:
            return GerritChangeRevisionFile.parse(
                result[0],
                change=self.change,
                revision=self.revision,
                gerrit=self.gerrit,
            )
        else:
            raise UnknownFile(path)

    def get(self, path):
        """
        get a file by path

        :param path: file path
        :return:
        """
        return self[path]



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.accounts.account import GerritAccount


class GerritAccounts(object):
    def __init__(self, gerrit):
        self.gerrit = gerrit
        self.endpoint = "/accounts"

    def search(
        self,
        query,
        limit=None,
        skip=None,
        detailed=False,
        suggested=False,
        all_emails=False,
    ):
        """
        Queries accounts visible to the caller.

        :param query: Query string
        :param limit: Int value that allows to limit the number of accounts
                      to be included in the output results
        :param skip: Int value that allows to skip the given
                     number of accounts from the beginning of the list
        :param detailed: boolean value, if True then full name,
                         preferred email, username and avatars for each account
                         will be added to the output result
        :param suggested: boolean value, if True get account suggestions
                          based on query string. If a result limit n is not
                          specified, then the default 10 is used.
        :param all_emails: boolean value, if True then all registered emails
                           for each account will be added to the output result
        :return:
        """
        option = filter(
            None,
            ["DETAILS" if detailed else None, "ALL_EMAILS" if all_emails else None],
        )
        option = None if not option else option
        params = {
            k: v for k, v in (("n", limit), ("S", skip), ("o", option)) if v is not None
        }

        endpoint = self.endpoint + "/?"
        if suggested:
            endpoint += "suggest&"
        endpoint += f"q={query}"

        return self.gerrit.get(endpoint, params=params)

    def get(self, username, detailed=False):
        """
        Returns an account

        :param username: username or _account_id
        :param detailed: boolean type, If True then fetch info in more details, such as:
            registered_on
        :return:
        """
        endpoint = self.endpoint + f"/{username}/"
        if detailed:
            endpoint += "detail"
        result = self.gerrit.get(endpoint)
        return GerritAccount(json=result, gerrit=self.gerrit)

    def create(self, username, input_):
        """
        Creates a new account.

        .. code-block:: python

            input_ = {
                "name": "John Doe",
                "email": "john.doe@example.com",
                "ssh_key": "ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAQEA0T...YImydZAw==",
                "http_password": "19D9aIn7zePb",
                "groups": [
                  "MyProject-Owners"
                ]
            }
            new_account = client.accounts.create('john.doe', input_)

        :param username: account username
        :param input_: the AccountInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html#account-input
        :return:
        """
        result = self.gerrit.put(
            self.endpoint + f"/{username}", json=input_, headers=self.gerrit.default_headers)
        return GerritAccount(json=result, gerrit=self.gerrit)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.projects.branches import GerritProjectBranches
from gerrit.projects.tags import GerritProjectTags
from gerrit.projects.commit import GerritProjectCommit
from gerrit.projects.dashboards import GerritProjectDashboards
from gerrit.projects.labels import GerritProjectLabels
from gerrit.projects.webhooks import GerritProjectWebHooks
from gerrit.changes.change import GerritChange
from gerrit.utils.models import BaseModel


class GerritProject(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = f"/projects/{self.id}"

    def get_description(self):
        """
        Retrieves the description of a project.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/description")

    def set_description(self, input_):
        """
        Sets the description of a project.

        .. code-block:: python

            input_ = {
                "description": "Plugin for Gerrit that handles the replication.",
                "commit_message": "Update the project description"
            }
            project = client.projects.get('myproject')
            result = project.set_description(input_)

        :param input_: the ProjectDescriptionInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #project-description-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/description",
                               json=input_, headers=self.gerrit.default_headers)

    def delete_description(self):
        """
        Deletes the description of a project.

        :return:
        """
        self.gerrit.delete(self.endpoint + "/description")

    def delete(self):
        """
        Delete the project, requires delete-project plugin

        :return:
        """
        self.gerrit.post(self.endpoint + "/delete-project~delete")

    def get_parent(self):
        """
        Retrieves the name of a project’s parent project. For the All-Projects root project an empty
        string is returned.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/parent")

    def set_parent(self, input_):
        """
        Sets the parent project for a project.

        .. code-block:: python

            input_ = {
                "parent": "Public-Plugins",
                "commit_message": "Update the project parent"
            }
            project = client.projects.get('myproject')
            result = project.set_parent(input_)

        :param input_: The ProjectParentInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #project-parent-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/parent",
                               json=input_, headers=self.gerrit.default_headers)

    def get_HEAD(self):
        """
        Retrieves for a project the name of the branch to which HEAD points.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/HEAD")

    def set_HEAD(self, input_):
        """
        Sets HEAD for a project.

        .. code-block:: python

            input_ = {
                "ref": "refs/heads/stable"
            }
            project = client.projects.get('myproject')
            result = project.set_HEAD(input_)

        :param input_: The HeadInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#head-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/HEAD",
                               json=input_, headers=self.gerrit.default_headers)

    def get_config(self):
        """
        Gets some configuration information about a project.
        Note that this config info is not simply the contents of project.config; it generally
        contains fields that may
        have been inherited from parent projects.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/config")

    def set_config(self, input_):
        """
        Sets the configuration of a project.

        .. code-block:: python

            input_ = {
                "description": "demo project",
                "use_contributor_agreements": "FALSE",
                "use_content_merge": "INHERIT",
                "use_signed_off_by": "INHERIT",
                "create_new_change_for_all_not_in_target": "INHERIT",
                "enable_signed_push": "INHERIT",
                "require_signed_push": "INHERIT",
                "reject_implicit_merges": "INHERIT",
                "require_change_id": "TRUE",
                "max_object_size_limit": "10m",
                "submit_type": "REBASE_IF_NECESSARY",
                "state": "ACTIVE"
            }
            project = client.projects.get('myproject')
            result = project.set_config(input_)

        :param input_: the ConfigInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#config-info
        :return:
        """
        return self.gerrit.put(self.endpoint + "/config",
                               json=input_, headers=self.gerrit.default_headers)

    def get_statistics(self):
        """
        Return statistics for the repository of a project.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/statistics.git")

    def run_garbage_collection(self, input_):
        """
        Run the Git garbage collection for the repository of a project.

        .. code-block:: python

            input_ = {
                "show_progress": "true"
            }
            project = client.projects.get('myproject')
            result = project.run_garbage_collection(input_)

        :param input_: the GCInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#gc-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/gc",
                                json=input_, headers=self.gerrit.default_headers)

    def ban_commits(self, input_):
        """
        Marks commits as banned for the project.

        .. code-block:: python

            input_ = {
                "commits": [
                  "a8a477efffbbf3b44169bb9a1d3a334cbbd9aa96",
                  "cf5b56541f84b8b57e16810b18daca9c3adc377b"
                ],
                "reason": "Violates IP"
            }
            project = client.projects.get('myproject')
            result = project.ban_commits(input_)

        :param input_: the BanInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#ban-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/ban",
                               json=input_, headers=self.gerrit.default_headers)

    def get_access_rights(self):
        """
        Lists the access rights for a single project.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/access")

    def set_access_rights(self, input_):
        """
        Sets access rights for the project using the diff schema provided by ProjectAccessInput.
        https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html#set-access

        :param input_: the ProjectAccessInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #project-access-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/access",
                                json=input_, headers=self.gerrit.default_headers)

    def create_change(self, input_):
        """
        Create Change for review. This endpoint is functionally equivalent to create change in the
        change API, but it has the project name in the URL, which is easier to route in sharded
        deployments.
        support this method since v3.3.0

        .. code-block:: python

            input_ = {
                "subject": "Let's support 100% Gerrit workflow direct in browser",
                "branch": "stable",
                "topic": "create-change-in-browser",
                "status": "NEW"
            }

            project = client.projects.get('myproject')
            result = project.create_change(input_)

        :param input_:
        :return:
        """
        result = self.gerrit.post(self.endpoint + "/create.change",
                                  json=input_, headers=self.gerrit.default_headers)
        return GerritChange(json=result, gerrit=self.gerrit)

    def create_access_rights_change(self, input_):
        """
        Sets access rights for the project using the diff schema provided by ProjectAccessInput
        This takes the same input as Update Access Rights, but creates a pending change for review.
        Like Create Change, it returns a ChangeInfo entity describing the resulting change.
        https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
        #create-access-change

        :param input_: the ProjectAccessInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #project-access-input
        :return:
        """
        result = self.gerrit.put(self.endpoint + "/access:review",
                                 json=input_, headers=self.gerrit.default_headers)
        return GerritChange(json=result, gerrit=self.gerrit)

    def check_access(self, options):
        """
        runs access checks for other users.

        :param options:
          Check Access Options
            * Account(account): The account for which to check access. Mandatory.
            * Permission(perm): The ref permission for which to check access.
              If not specified, read access to at least branch is checked.
            * Ref(ref): The branch for which to check access. This must be given if perm is
            specified.

        :return:
        """
        return self.gerrit.get(self.endpoint + f"/check.access?{options}")

    def index(self, input_):
        """
        Adds or updates the current project (and children, if specified) in the secondary index.
        The indexing task is executed asynchronously in background and this command returns
        immediately if async is specified in the input.

        .. code-block:: python

            input_ = {
                "index_children": "true"
                "async": "true"
            }
            project = client.projects.get('myproject')
            result = project.index(input_)

        :param input_: the IndexProjectInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #index-project-input
        :return:
        """
        self.gerrit.post(self.endpoint + "/index",
                         json=input_, headers=self.gerrit.default_headers)

    def index_all_changes(self):
        """
        Adds or updates the current project (and children, if specified) in the secondary index.
        The indexing task is executed asynchronously in background and this command returns
        immediately if async is specified in the input.

        :return:
        """
        self.gerrit.post(self.endpoint + "/index.changes")

    def check_consistency(self, input_):
        """
        Performs consistency checks on the project.

        .. code-block:: python

            input_ = {
                "auto_closeable_changes_check": {
                    "fix": 'true',
                    "branch": "refs/heads/master",
                    "max_commits": 100
                }
            }

            project = client.projects.get('myproject')
            result = project.check_consistency(input_)

        :param input_: the CheckProjectInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #check-project-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/check",
                                json=input_, headers=self.gerrit.default_headers)

    @property
    def branches(self):
        """
        List the branches of a project. except the refs/meta/config

        :return:
        """
        return GerritProjectBranches(self.id, self.gerrit)

    @property
    def child_projects(self):
        """
        List the direct child projects of a project.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/children/")

    @property
    def tags(self):
        """
        List the tags of a project.

        :return:
        """
        return GerritProjectTags(self.id, self.gerrit)

    def get_commit(self, commit):
        """
        Retrieves a commit of a project.

        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/commits/{commit}")
        return GerritProjectCommit(json=result, project=self.id, gerrit=self.gerrit)

    @property
    def dashboards(self):
        """
        gerrit dashboards operations

        :return:
        """
        return GerritProjectDashboards(project=self.id, gerrit=self.gerrit)

    def get_labels(self):
        """
        gerrit labels or gerrit labels operations

        :return:
        """
        version = self.gerrit.version
        if parse(version) < parse("3.2.0"):
            result = self.gerrit.get(self.endpoint)
            return result.get("labels")
        else:
            return GerritProjectLabels(project=self.id, gerrit=self.gerrit)

    @property
    def webhooks(self):
        """
        gerrit webhooks operations, requires webhooks plugin

        :return:
        """
        return GerritProjectWebHooks(project=self.id, gerrit=self.gerrit)



"""Module for custom exceptions.

Where possible we try to throw exceptions with non-generic,
meaningful names.
"""


class GerritAPIException(Exception):
    """
    Base class for all errors
    """


class ClientError(GerritAPIException):
    """
    ClientError
    """


class ServerError(GerritAPIException):
    """
    ServerError
    """


class UnauthorizedError(GerritAPIException):
    """
    401 Unauthorized
    """


class AuthError(ClientError):
    """
    403 Forbidden is returned if the operation is not allowed because the calling user does not have
    sufficient permissions.
    """


class ValidationError(ClientError):
    """
    400 Bad Request is returned if the request is not understood by the server due to malformed
    syntax.
    E.g. 400 Bad Request is returned if JSON input is expected but the 'Content-Type' of the request
     is not 'application/json' or the request body doesn't contain valid JSON.
    400 Bad Request is also returned if required input fields are not set or if options are set
    which cannot be used together.
    """


class NotAllowedError(ClientError):
    """
    405 Method Not Allowed is returned if the resource exists but doesn't support the operation.
    """


class ConflictError(ClientError):
    """
    409 Conflict is returned if the request cannot be completed because the current state of the
    resource doesn't allow the operation.
    """


class NotFoundError(ClientError):
    """
    Resource cannot be found
    """


class UnknownBranch(KeyError, NotFoundError):
    """
    Gerrit does not recognize the branch requested.
    """


class UnknownTag(KeyError, NotFoundError):
    """
    Gerrit does not recognize the tag requested.
    """


class UnknownFile(KeyError, NotFoundError):
    """
    Gerrit does not recognize the revision file requested.
    """



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi

from gerrit.utils.models import BaseModel


class GerritChangeRevisionComment(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = f"/changes/{self.change}/revisions/{self.revision}/comments/{self.id}"

    def delete(self, input_=None):
        """
        Deletes a published comment of a revision. Instead of deleting the whole comment,
        this endpoint just replaces the comment’s message with a new message, which contains
        the name of the user who deletes the comment and the reason why it’s deleted.

        .. code-block:: python

            input_ = {
                "reason": "contains confidential information"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision('3848807f587dbd3a7e61723bbfbf1ad13ad5a00a')
            comment = revision.comments.get("e167e775_e069567a")
            result = comment.delete(input_)
            # or
            result = comment.delete()

        :param input_: the DeleteCommentInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #delete-comment-input
        :return:
        """
        if input_ is None:
            return self.gerrit.delete(self.endpoint)
        else:
            return self.gerrit.post(self.endpoint + "/delete",
                                    json=input_, headers=self.gerrit.default_headers)


class GerritChangeRevisionComments(object):
    def __init__(self, change, revision, gerrit):
        self.change = change
        self.revision = revision
        self.gerrit = gerrit
        self.endpoint = f"/changes/{self.change}/revisions/{self.revision}/comments"

    def list(self):
        """
        Lists the published comments of a revision.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        comments = []
        for key, value in result.items():
            for item in value:
                comment = item
                comment.update({"path": key})
                comments.append(comment)
        return GerritChangeRevisionComment.parse_list(
            comments, change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    def get(self, id_):
        """
        Retrieves a published comment of a revision.

        :param id_:
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{id_}")
        return GerritChangeRevisionComment.parse(
            result, change=self.change, revision=self.revision, gerrit=self.gerrit
        )



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
import json
from gerrit.utils.common import logger


class Entity(object):
    required = ()
    optional = ()

    def __init__(self, **kwargs):
        """

        :param kwargs:
        """
        self.attributes = self.required + self.optional

        for item in self.required:
            if item not in kwargs.keys():
                name = self.__class__.__name__
                logger.warning(
                    f"*** Warning! {name} missing a required keyword argument '{item}' ***"
                )

        for (k, v) in kwargs.items():
            if k in self.attributes:
                setattr(self, k, v)
            else:
                name = self.__class__.__name__
                logger.warning(
                    f"*** Warning! {name} got an unexpected keyword argument: '{k}' ***"
                )

    def __getattr__(self, key):
        return self.__dict__.get(key, None)

    def __str__(self):
        """

        :return:
        """
        review_input = {}
        for key in self.attributes:
            value = getattr(self, key)
            if value:
                review_input.update({key: value})

        return json.dumps(review_input, sort_keys=True)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from base64 import b64decode
from gerrit.changes.revision.drafts import GerritChangeRevisionDrafts
from gerrit.changes.revision.comments import GerritChangeRevisionComments
from gerrit.changes.revision.files import GerritChangeRevisionFiles


class GerritChangeRevision(object):
    def __init__(self, gerrit, project, change, revision="current"):
        self.project = project
        self.change = change
        self.revision = revision
        self.gerrit = gerrit
        self.endpoint = f"/changes/{self.change}/revisions/{self.revision}"

    def get_commit(self):
        """
        Retrieves a parsed commit of a revision.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/commit")
        commit = result.get("commit")
        if commit:
            project = self.gerrit.projects.get(self.project)
            if project:
                return project.get_commit(commit)

    def get_description(self):
        """
        Retrieves the description of a patch set.
        If the patch set does not have a description an empty string is returned.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/description")

    def set_description(self, input_):
        """
        Sets the description of a patch set.

        .. code-block:: python

            input_ = {
                "description": "Added Documentation"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision('3848807f587dbd3a7e61723bbfbf1ad13ad5a00a')
            result = revision.set_description(input_)

        :param input_: the DescriptionInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #description-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/description",
                               json=input_, headers=self.gerrit.default_headers)

    def get_merge_list(self):
        """
        Returns the list of commits that are being integrated into a target branch by a merge
        commit. By default, the first parent is assumed to be uninteresting. By using the parent
        option another parent can be set as uninteresting (parents are 1-based).

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/mergelist")
        return [
            self.gerrit.projects.get(self.project).get_commit(item.get("commit"))
            for item in result
        ]

    def get_revision_actions(self):
        """
        Retrieves revision actions of the revision of a change.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/actions")

    def get_review(self):
        """
        Retrieves a review of a revision.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/review")

    def get_related_changes(self):
        """
        Retrieves related changes of a revision. Related changes are changes that either depend on,
        or are dependencies of the revision.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/related")

    def set_review(self, input_):
        """
        Sets a review on a revision, optionally also publishing draft comments, setting labels,
        adding reviewers or CCs, and modifying the work in progress property.
        A review cannot be set on a change edit. Trying to post a review for a change edit fails
        with 409 Conflict.

        .. code-block:: python

            input_ = {
                "tag": "jenkins",
                "message": "Some nits need to be fixed.",
                "labels": {
                    "Code-Review": -1
                },
                "comments": {
                      "sonarqube/cloud/project_badges.py": [
                            {
                                "line": 23,
                                "message": "[nit] trailing whitespace"
                            },
                            {
                                "line": 49,
                                "message": "[nit] s/conrtol/control"
                            },
                            {
                                "range": {
                                    "start_line": 50,
                                    "start_character": 0,
                                    "end_line": 55,
                                    "end_character": 20
                                },
                                "message": "Incorrect indentation"
                            }
                      ]
                }
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision('3848807f587dbd3a7e61723bbfbf1ad13ad5a00a')
            result = revision.set_review(input_)

        :param input_: the ReviewInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#review-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/review",
                                json=input_, headers=self.gerrit.default_headers)

    def rebase(self, input_):
        """
        Rebases a revision.
        Optionally, the parent revision can be changed to another patch set through the RebaseInput
        entity.

        .. code-block:: python

            input_ = {
                "base" : "1234"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision('3848807f587dbd3a7e61723bbfbf1ad13ad5a00a')
            result = revision.rebase(input_)

        :param input_: the RebaseInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#rebase-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/rebase",
                                json=input_, headers=self.gerrit.default_headers)

    def submit(self):
        """
        Submits a revision.
        If the revision cannot be submitted, e.g. because the submit rule doesn’t allow submitting
        the revision or the revision is not the current revision, the response is 409 Conflict and
        the error message is contained in the response body.

        :return:
        """
        return self.gerrit.post(self.endpoint + "/submit")

    def get_patch(self, zip_=False, download=False, path=None, decode=False):
        """
        Gets the formatted patch for one revision.
        The formatted patch is returned as text encoded inside base64 if decode is False.

        Adding query parameter zip (for example /changes/…​/patch?zip) returns the patch as a
        single file inside of a ZIP archive. Clients can expand the ZIP to obtain the plain text
        patch, avoiding the need for a base64 decoding step. This option implies download.

        Query parameter download (e.g. /changes/…​/patch?download) will suggest the browser
        save the patch as commitsha1.diff.base64, for later processing by command line tools.

        If the path parameter is set, the returned content is a diff of the single file that
        the path refers to.

        :param zip_:
        :param download:
        :param path:
        :param decode: Decode bas64 to plain text.
        :return:
        """
        endpoint = self.endpoint + "/patch"

        if zip_:
            endpoint += endpoint + "?zip"

        if download:
            endpoint += endpoint + "?download"

        if path:
            endpoint += endpoint + f"?path={quote_plus(path)}"

        result = self.gerrit.get(endpoint)
        if decode:
            return b64decode(result).decode("utf-8")
        return result

    def submit_preview(self):
        """
        need fix bug
        Gets a file containing thin bundles of all modified projects if this change was submitted.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/preview_submit")

    def is_mergeable(self):
        """
        Gets the method the server will use to submit (merge) the change and an indicator
        if the change is currently mergeable.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/mergeable")

    def get_submit_type(self):
        """
        Gets the method the server will use to submit (merge) the change.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/submit_type")

    def test_submit_type(self, input_):
        """
        Tests the submit_type Prolog rule in the project, or the one given.

        :param input_: the Prolog code
        :type: str
        :return:
        """
        return self.gerrit.post(self.endpoint + "/test.submit_type",
                                data=input_, headers={"Content-Type": "plain/text"})

    def test_submit_rule(self, input_):
        """
        Tests the submit_rule Prolog rule in the project, or the one given.

        :param input_: the Prolog code
        :type: str
        :return:
        """
        return self.gerrit.post(self.endpoint + "/test.submit_rule",
                                data=input_, headers={"Content-Type": "plain/text"})

    @property
    def drafts(self):
        return GerritChangeRevisionDrafts(
            change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    @property
    def comments(self):
        return GerritChangeRevisionComments(
            change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    def list_robot_comments(self):
        """
        Lists the robot comments of a revision.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/robotcomments")

    def get_robot_comment(self, commit_id):
        """
        Retrieves a robot comment of a revision.

        :param commit_id:
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/robotcomments/{commit_id}")

    @property
    def files(self):
        return GerritChangeRevisionFiles(
            change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    def cherry_pick(self, input_):
        """
        Cherry picks a revision to a destination branch.

        .. code-block:: python

            input_ = {
                "message" : "Implementing Feature X",
                "destination" : "release-branch"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision('3848807f587dbd3a7e61723bbfbf1ad13ad5a00a')
            result = revision.cherry_pick(input_)

        :param input_: the CherryPickInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #cherry-pick-commit
        :return:
        """
        return self.gerrit.post(self.endpoint + "/cherrypick",
                                json=input_, headers=self.gerrit.default_headers)

    def list_reviewers(self):
        """
        Lists the reviewers of a revision.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/reviewers")

    def list_votes(self, account):
        """
        Lists the votes for a specific reviewer of the revision.

        :param account: account id or username
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/reviewers/{account}/votes/")

    def delete_vote(self, account, label, input_=None):
        """
        Deletes a single vote from a revision. The deletion will be possible only
        if the revision is the current revision. By using this endpoint you can prevent
        deleting the vote (with same label) from a newer patch set by mistake.
        Note, that even when the last vote of a reviewer is removed the reviewer itself is
        still listed on the change.

        .. code-block:: python

            input_ = {
                "notify": "NONE"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision("0f4f97b5af9a965e082fb8cde082c5f1ba2fe930")
            revision.delete_vote('John', 'Code-Review', input_)
            # or
            revision.delete_vote('John', 'Code-Review')

        :param account:
        :param label:
        :param input_: the DeleteVoteInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #delete-vote-input
        :return:
        """
        endpoint = self.endpoint + f"/reviewers/{account}/votes/{label}"
        if input_ is None:
            self.gerrit.delete(endpoint)
        else:
            endpoint += "/delete"
            self.gerrit.post(endpoint, json=input_, headers=self.gerrit.default_headers)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel


class GerritAccountGPGKey(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = f"/accounts/{self.username}/gpgkeys"

    def delete(self):
        """
        Deletes a GPG key of a user.

        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{self.id}")


class GerritAccountGPGKeys(object):
    def __init__(self, username, gerrit):
        self.username = username
        self.gerrit = gerrit
        self.endpoint = f"/accounts/{self.username}/gpgkeys"

    def list(self):
        """
        Returns the GPG keys of an account.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        keys = []
        for key, value in result.items():
            gpg_key = value
            gpg_key.update({"id": key})
            keys.append(gpg_key)

        return GerritAccountGPGKey.parse_list(
            keys, username=self.username, gerrit=self.gerrit
        )

    def get(self, id_):
        """
        Retrieves a GPG key of a user.

        :param id_: GPG key id
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{id_}")
        return GerritAccountGPGKey(json=result, username=self.username, gerrit=self.gerrit)

    def modify(self, input_):
        """
        Add or delete one or more GPG keys for a user.

        .. code-block:: python

            input_ = {
                "add": [
                  "-----BEGIN PGP PUBLIC KEY BLOCK-----\\n
                  Version: GnuPG v1\\n\\n
                  mQENBFXUpNcBCACv4paCiyKxZ0EcKy8VaWVNkJlNebRBiyw9WxU85wPOq5Gz/3GT\\n
                  RQwKqeY0SxVdQT8VNBw2sBe2m6eqcfZ2iKmesSlbXMe15DA7k8Bg4zEpQ0tXNG1L\\n
                  hceZDVQ1Xk06T2sgkunaiPsXi82nwN3UWYtDXxX4is5e6xBNL48Jgz4lbqo6+8D5\\n
                  vsVYiYMx4AwRkJyt/oA3IZAtSlY8Yd445nY14VPcnsGRwGWTLyZv9gxKHRUppVhQ\\n
                  E3o6ePXKEVgmONnQ4CjqmkGwWZvjMF2EPtAxvQLAuFa8Hqtkq5cgfgVkv/Vrcln4\\n
                  nQZVoMm3a3f5ODii2tQzNh6+7LL1bpqAmVEtABEBAAG0H0pvaG4gRG9lIDxqb2hu\\n
                  LmRvZUBleGFtcGxlLmNvbT6JATgEEwECACIFAlXUpNcCGwMGCwkIBwMCBhUIAgkK\\n
                  CwQWAgMBAh4BAheAAAoJEJNQnkuvyKSbfjoH/2OcSQOu1kJ20ndjhgY2yNChm7gd\\n
                  tU7TEBbB0TsLeazkrrLtKvrpW5+CRe07ZAG9HOtp3DikwAyrhSxhlYgVsQDhgB8q\\n
                  G0tYiZtQ88YyYrncCQ4hwknrcWXVW9bK3V4ZauxzPv3ADSloyR9tMURw5iHCIeL5\\n
                  fIw/pLvA3RjPMx4Sfow/bqRCUELua39prGw5Tv8a2ZRFbj2sgP5j8lUFegyJPQ4z\\n
                  tJhe6zZvKOzvIyxHO8llLmdrImsXRL9eqroWGs0VYqe6baQpY6xpSjbYK0J5HYcg\\n
                  TO+/u80JI+ROTMHE6unGp5Pgh/xIz6Wd34E0lWL1eOyNfGiPLyRWn1d0yZO5AQ0E\\n
                  VdSk1wEIALUycrH2HK9zQYdR/KJo1yJJuaextLWsYYn881yDQo/p06U5vXOZ28lG\\n
                  Aq/Xs96woVZPbgME6FyQzhf20Z2sbr+5bNo3OcEKaKX3Eo/sWwSJ7bXbGLDxMf4S\\n
                  etfY1WDC+4rTqE30JuC++nQviPRdCcZf0AEgM6TxVhYEMVYwV787YO1IH62EBICM\\n
                  SkIONOfnusNZ4Skgjq9OzakOOpROZ4tki5cH/5oSDgdcaGPy1CFDpL9fG6er2zzk\\n
                  sw3qCbraqZrrlgpinWcAduiao67U/dV18O6OjYzrt33fTKZ0+bXhk1h1gloC21MQ\\n
                  ya0CXlnfR/FOQhvuK0RlbR3cMfhZQscAEQEAAYkBHwQYAQIACQUCVdSk1wIbDAAK\\n
                  CRCTUJ5Lr8ikm8+QB/4uE+AlvFQFh9W8koPdfk7CJF7wdgZZ2NDtktvLL71WuMK8\\n
                  POmf9f5JtcLCX4iJxGzcWogAR5ed20NgUoHUg7jn9Xm3fvP+kiqL6WqPhjazd89h\\n
                  k06v9hPE65kp4wb0fQqDrtWfP1lFGuh77rQgISt3Y4QutDl49vXS183JAfGPxFxx\\n
                  8FgGcfNwL2LVObvqCA0WLqeIrQVbniBPFGocE3yA/0W9BB/xtolpKfgMMsqGRMeu\\n
                  9oIsNxB2oE61OsqjUtGsnKQi8k5CZbhJaql4S89vwS+efK0R+mo+0N55b0XxRlCS\\n
                  faURgAcjarQzJnG0hUps2GNO/+nM7UyyJAGfHlh5\\n
                  =EdXO\\n
                  -----END PGP PUBLIC KEY BLOCK-----\\n"
                ],
                "delete": [
                  "DEADBEEF",
                ]
            }
            account = client.accounts.get('kevin.shi')
            result = account.gpg_keys.modify(input_)

        :param input_: the GpgKeysInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html#gpg-keys-input
        :return:
        """
        return self.gerrit.post(self.endpoint , json=input_, headers=self.gerrit.default_headers)

    def delete(self, id_):
        """
        Deletes a GPG key of a user.

        :param id_: GPG key id
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{id_}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.changes.reviewers import GerritChangeReviewers
from gerrit.changes.revision import GerritChangeRevision
from gerrit.changes.edit import GerritChangeEdit
from gerrit.changes.messages import GerritChangeMessages
from gerrit.utils.models import BaseModel


class GerritChange(BaseModel):
    def __init__(self, **kwargs):
        self.revisions = {}
        self.current_revision_number = 0
        super().__init__(**kwargs)
        self.endpoint = f"/changes/{self.id}"

    def get_meta_diff(self, old=None, meta=None):
        """
        Retrieves the difference between two historical states of a change by
        specifying the and the parameters. old=SHA-1,meta=SHA-1.
        If the parameter is not provided, the parent of the SHA-1 is used.
        If the parameter is not provided, the current state of the change is used.
        If neither are provided, the difference between the current state of the change and
        its previous state is returned.

        :param old:
        :param meta:
        :return:
        """
        params = {}
        if old:
            params.update({"old": old})

        if meta:
            params.update({"meta": meta})

        return self.gerrit.get(self.endpoint + "/meta_diff", params=params)

    def create_merge_patch_set(self, input_):
        """
        Update an existing change by using a MergePatchSetInput entity.
        Gerrit will create a merge commit based on the information of MergePatchSetInput and add
        a new patch set to the change corresponding to the new merge commit.

        .. code-block:: python

            input_ = {
                "subject": "Merge master into stable",
                "merge": {
                  "source": "refs/heads/master"
                }
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.update(input_)

        :param input_: the MergePatchSetInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #merge-patch-set-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/merge",
                                json=input_, headers=self.gerrit.default_headers)

    def set_commit_message(self, input_):
        """
        Creates a new patch set with a new commit message.

        .. code-block:: python

            input_ = {
                "message": "New Commit message \\n\\nChange-Id: I10394472cbd17dd12454f22b143a444\\n"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.set_commit_message(input_)

        :param input_: the CommitMessageInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #commit-message-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/message",
                               json=input_, headers=self.gerrit.default_headers)

    def list_votes(self, account):
        """
        Lists the votes for a specific reviewer of the change.

        :param account: account id or username
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/reviewers/{account}/votes")

    def delete_vote(self, account, label, input_=None):
        """
        Deletes a single vote from a change. Note, that even when the last vote of a reviewer is
        removed the reviewer itself is still listed on the change.
        If another user removed a user’s vote, the user with the deleted vote will be added to
        the attention set.

        .. code-block:: python

            input_ = {
                "notify": "NONE"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            change.delete_vote('John', 'Code-Review', input_)
            # or
            change.delete_vote('John', 'Code-Review')

        :param account:
        :param label:
        :param input_: the DeleteVoteInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #delete-vote-input
        :return:
        """
        endpoint = f"/changes/{self.change}/reviewers/{account}/votes/{label}"
        if input_ is None:
            self.gerrit.delete(endpoint)
        else:
            endpoint += "/delete"
            self.gerrit.post(endpoint, json=input_, headers=self.gerrit.default_headers)

    def get_topic(self):
        """
        Retrieves the topic of a change.

        :getter: Retrieves the topic of a change.
        :setter: Sets the topic of a change.
        :deleter: Deletes the topic of a change.

        :return:
        """
        return self.gerrit.get(f"/changes/{self.id}/topic")

    def set_topic(self, topic):
        """
        Sets the topic of a change.

        :param topic: The new topic
        :return:
        """
        input_ = {"topic": topic}
        return self.gerrit.put(self.endpoint + "/topic",
                               json=input_, headers=self.gerrit.default_headers)

    def delete_topic(self):
        """
        Deletes the topic of a change.

        :return:
        """
        self.gerrit.delete(f"/changes/{self.id}/topic")

    def get_assignee(self):
        """
        Retrieves the account of the user assigned to a change.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/assignee")
        if result:
            username = result.get("username")
            return self.gerrit.accounts.get(username)

    def set_assignee(self, input_):
        """
        Sets the assignee of a change.

        .. code-block:: python

            input_ = {
                "assignee": "jhon.doe"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.set_assignee(input_)

        :param input_: the AssigneeInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#assignee-input
        :return:
        """
        result = self.gerrit.put(self.endpoint + "/assignee",
                                 json=input_, headers=self.gerrit.default_headers)
        if result:
            username = result.get("username")
            return self.gerrit.accounts.get(username)

    def get_past_assignees(self):
        """
        Returns a list of every user ever assigned to a change, in the order in which they were
        first assigned.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/past_assignees")
        assignees = []
        if result:
            for item in result:
                username = item.get("username")
                assignee = self.gerrit.accounts.get(username)
                assignees.append(assignee)

        return assignees

    def delete_assignee(self):
        """
        Deletes the assignee of a change.

        :return:
        """
        response = self.gerrit.delete(self.endpoint + "/assignee")
        result = self.gerrit.decode_response(response)

        if result:
            username = result.get("username")
            return self.gerrit.accounts.get(username)

    def get_pure_revert(self, commit):
        """
        Check if the given change is a pure revert of the change it references in revertOf.

        :param commit: commit id
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/pure_revert?o={commit}")

    def abandon(self):
        """
        Abandons a change.
        Abandoning a change also removes all users from the attention set.
        If the change cannot be abandoned because the change state doesn't allow abandoning of
        the change, the response is “409 Conflict” and the error message is contained in the
        response body.

        :return:
        """
        return self.gerrit.post(self.endpoint + "/abandon")

    def restore(self):
        """
        Restores a change.
        If the change cannot be restored because the change state doesn't allow restoring the
        change, the response is “409 Conflict” and the error message is contained in the
        response body.

        :return:
        """
        return self.gerrit.post(self.endpoint + "/restore")

    def rebase(self, input_):
        """
        Rebase a change.
        If the change cannot be rebased, e.g. due to conflicts, the response is '409 Conflict'
        and the error message is contained in the response body.

        .. code-block:: python

            input_ = {
                "base" : "1234",
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.rebase(input_)

        :param input_: the RebaseInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#rebase-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/rebase",
                                json=input_, headers=self.gerrit.default_headers)

    def move(self, input_):
        """
        Move a change.
        If the change cannot be moved because the change state doesn't allow moving the change,
        the response is '409 Conflict' and the error message is contained in the response body.

        .. code-block:: python

            input_ = {
                "destination_branch" : "release-branch"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.move(input_)

        :param input_: the MoveInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#move-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/move",
                                json=input_,
                                headers=self.gerrit.default_headers)

    def revert(self, input_=None):
        """
        Reverts a change.
        The request body does not need to include a RevertInput entity if no review comment is
        added.

        If the user doesn't have revert permission on the change or upload permission on the
        destination branch, the response is '403 Forbidden', and the error message is contained in
        the response body.

        If the change cannot be reverted because the change state doesn't allow reverting the
        change, the response is 409 Conflict and the error message is contained in the
        response body.

        .. code-block:: python

            input_ = {
                "message" : "Message to be added as review comment to the change when reverting the
                change."
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.revert()
            # or
            result = change.revert(input_)

        :param input_: the RevertInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#revert-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/revert",
                                json=input_ or {}, headers=self.gerrit.default_headers)

    def revert_submission(self):
        """
        Creates open revert changes for all of the changes of a certain submission.

        If the user doesn't have revert permission on the change or upload permission on the
        destination, the response is '403 Forbidden', and the error message is contained in the
        response body.

        If the change cannot be reverted because the change state doesn't allow reverting the change
        the response is '409 Conflict', and the error message is contained in the response body.

        :return:
        """
        return self.gerrit.post(self.endpoint + "/revert_submission")

    def submit(self, input_=None):
        """
        Submits  a change.
        Submitting a change also removes all users from the attention set.

        If the change cannot be submitted because the submit rule doesn't allow submitting
        the change, the response is 409 Conflict and the error message is contained in the
        response body.

        .. code-block:: python

            input_ = {
                "on_behalf_of": 1001439
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.submit(input_)

        :param input_: the SubmitInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#submit-input
        :return:
        """
        endpoint = self.endpoint + "/submit"
        if input_ is None:
            result = self.gerrit.post(endpoint)
        else:
            result = self.gerrit.post(endpoint,
                                    json=input_, headers=self.gerrit.default_headers)
        return result

    def list_submitted_together_changes(self):
        """
        Computes list of all changes which are submitted when Submit is called for this change,
        including the current change itself.

        """
        return self.gerrit.get(self.endpoint + "/submitted_together?o=NON_VISIBLE_CHANGES")

    def delete(self):
        """
        Deletes a change.

        :return:
        """
        self.gerrit.delete(self.endpoint)

    def get_include_in(self):
        """
        Retrieves the branches and tags in which a change is included.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/in")

    def index(self):
        """
        Adds or updates the change in the secondary index.

        :return:
        """
        self.gerrit.post(self.endpoint + "/index")

    def list_comments(self):
        """
        Lists the published comments of all revisions of the change.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/comments")

    def list_robot_comments(self):
        """
        Lists the robot comments of all revisions of the change.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/robotcomments")

    def list_drafts(self):
        """
        Lists the draft comments of all revisions of the change that belong to the calling user.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/drafts")

    def consistency_check(self):
        """
        Performs consistency checks on the change, and returns a ChangeInfo entity with the problems
        field set to a list of ProblemInfo entities.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/check")

    def fix(self, input_=None):
        """
        Performs consistency checks on the change as with GET /check,
        and additionally fixes any problems that can be fixed automatically. The returned field
        values reflect any fixes.
        Some fixes have options controlling their behavior, which can be set in the FixInput
        entity body. Only the change owner, a project owner, or an administrator may fix changes.

        .. code-block:: python

            input_ = {
                "delete_patch_set_if_commit_missing": "true",
                "expect_merged_as": "something"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.fix()
            # or
            result = change.fix(input_)

        :param input_: the FixInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#fix-input
        :return:
        """
        endpoint = self.endpoint + "/check"
        if input_ is None:
            result = self.gerrit.post(endpoint)
        else:
            result = self.gerrit.post(endpoint, json=input_, headers=self.gerrit.default_headers)
        return result

    def set_work_in_progress(self, input_=None):
        """
        Marks the change as not ready for review yet.
        Changes may only be marked not ready by the owner, project owners or site administrators.
        Marking a change work in progress also removes all users from the attention set.

        The request body does not need to include a WorkInProgressInput entity if no review comment
        is added.

        .. code-block:: python

            input_ = {
                "message": "Refactoring needs to be done before we can proceed here."
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.set_work_in_progress(input_)
            # or
            result = change.set_work_in_progress()

        :param input_: the WorkInProgressInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #work-in-progress-input
        :return:
        """
        self.gerrit.post(self.endpoint + "/wip",
                         json=input_ or {}, headers=self.gerrit.default_headers
                         )

    def set_ready_for_review(self, input_):
        """
        Marks the change as ready for review (set WIP property to false).
        Changes may only be marked ready by the owner, project owners or site administrators.
        Marking a change ready for review also adds all of the reviewers of the change to the
        attention set.

        .. code-block:: python

            input_ = {
                'message': 'Refactoring is done.'
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            change.set_ready_for_review(input_)

        :param input_: the WorkInProgressInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #work-in-progress-input
        :return:
        """
        self.gerrit.post(self.endpoint + "/ready", json=input_, headers=self.gerrit.default_headers)

    def mark_private(self, input_):
        """
        Marks the change to be private. Only open changes can be marked private.
        Changes may only be marked private by the owner or site administrators.

        .. code-block:: python

            input_ = {
                "message": "After this security fix has been released we can make it public now."
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            change.mark_private(input_)

        :param input_: the PrivateInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#private-input
        :return:
        """
        self.gerrit.post(self.endpoint + "/private",
                         json=input_, headers=self.gerrit.default_headers
                         )

    def unmark_private(self, input_=None):
        """
        Marks the change to be non-private. Note users can only unmark own private changes.
        If the change was already not private, the response is '409 Conflict'.

        .. code-block:: python

            input_ = {
                "message": "This is a security fix that must not be public."
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            change.unmark_private(input_)
            # or
            change.unmark_private()

        :param input_: the PrivateInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#private-input
        :return:
        """
        if input_ is None:
            self.gerrit.delete(self.endpoint + "/private")
        else:
            self.gerrit.post(self.endpoint + "/private.delete",
                             json=input_, headers=self.gerrit.default_headers
                             )

    def ignore(self):
        """
        Marks a change as ignored. The change will not be shown in the incoming reviews' dashboard,
        and email notifications will be suppressed. Ignoring a change does not cause the change’s
        "updated" timestamp to be modified, and the owner is not notified.

        :return:
        """
        self.gerrit.put(self.endpoint + "/ignore")

    def unignore(self):
        """
        Un-marks a change as ignored.

        :return:
        """
        self.gerrit.put(self.endpoint + "/unignore")

    def mark_as_reviewed(self):
        """
        Marks a change as reviewed.

        :return:
        """
        self.gerrit.put(self.endpoint + "/reviewed")

    def mark_as_unreviewed(self):
        """
        Marks a change as unreviewed.

        :return:
        """
        self.gerrit.put(self.endpoint + "/unreviewed")

    def get_hashtags(self):
        """
        Gets the hashtags associated with a change.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/hashtags")

    def set_hashtags(self, input_):
        """
        Adds and/or removes hashtags from a change.

        .. code-block:: python

            input_ = {
                "add" : [
                    "hashtag3"
                ],
                "remove" : [
                    "hashtag2"
                ]
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.set_hashtags(input_)

        :param input_: the HashtagsInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#hashtags-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/hashtags",
                                json=input_, headers=self.gerrit.default_headers)


    @property
    def messages(self):
        return GerritChangeMessages(change=self.id, gerrit=self.gerrit)

    def check_submit_requirement(self, input_):
        """
        Tests a submit requirement.

        :param input_: the SubmitRequirementInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#submit-requirement-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/check.submit_requirement",
                                json=input_, headers=self.gerrit.default_headers)

    def get_edit(self):
        """
        Retrieves a change edit details.
        As response an EditInfo entity is returned that describes the change edit,
        or 204 No Content when change edit doesn't exist for this change.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/edit")
        return GerritChangeEdit(json=result, change=self.id, gerrit=self.gerrit)

    def create_empty_edit(self):
        """
        Creates empty change edit

        :return:
        """
        self.gerrit.post(self.endpoint + "/edit")

    @property
    def reviewers(self):
        return GerritChangeReviewers(change=self.id, gerrit=self.gerrit)

    def __get_revisions(self):
        endpoint = f"/changes/?q={self.number}&o=ALL_REVISIONS"
        [result] = self.gerrit.get(endpoint)
        self.revisions = {}
        for revision_sha, revision in result["revisions"].items():
            if result["current_revision"] == revision_sha:
                self.current_revision_number = revision["_number"]
            self.revisions[revision["_number"]] = revision_sha

    def __revision_number_to_sha(self, number):
        if number in self.revisions:
            return self.revisions[number]
        return None

    def get_revision(self, revision_id="current"):
        """
        Get one revision by revision SHA or integer number.

        :param revision_id: Optional ID. If not specified, the current revision will be retrieved.
                            It supports SHA IDs and integer numbers from -X to +X, where X is the
                            current (latest) revision.
                            Zero means current revision.
                            -N means the current revision number X minus N, so if the current
                            revision is 50, and -1 is given, the revision 49 will be retrieved.
        :return:
        """
        if isinstance(revision_id, int):
            if len(self.revisions) == 0:
                self.__get_revisions()
            if revision_id <= 0:
                revision_id = self.current_revision_number + revision_id
            revision_id = self.__revision_number_to_sha(revision_id)
            if revision_id is None:
                return None

        return GerritChangeRevision(
            gerrit=self.gerrit,
            project=self.project,
            change=self.id,
            revision=revision_id
        )

    def get_attention_set(self):
        """
        Returns all users that are currently in the attention set.
        support this method since v3.3.0

        :return:
        """
        return self.gerrit.get(f"/changes/{self.id}/attention")

    def add_to_attention_set(self, input_):
        """
        Adds a single user to the attention set of a change.
        support this method since v3.3.0

        A user can only be added if they are not in the attention set.
        If a user is added while already in the attention set, the request is silently ignored.

        .. code-block:: python

            input_ = {
                "user": "John Doe",
                "reason": "reason"
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            result = change.add_to_attention_set(input_)

        :param input_: the AttentionSetInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #attention-set-input
        :return:
        """
        result = self.gerrit.post(self.endpoint + "/attention",
                                  json=input_, headers=self.gerrit.default_headers)
        return result

    def remove_from_attention_set(self, id_, input_=None):
        """
        Deletes a single user from the attention set of a change.
        support this method since v3.3.0

        A user can only be removed from the attention set.
        if they are currently in the attention set. Otherwise, the request is silently ignored.

        .. code-block:: python

            input_ = {
                "reason": "reason"
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            change.remove_from_attention_set('kevin.shi', input_)
            # or
            change.remove_from_attention_set('kevin.shi')

        :param id_: account id
        :param input_: the AttentionSetInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #attention-set-input
        :return:
        """
        endpoint = self.endpoint + f"/attention/{id_}"
        if input_ is None:
            self.gerrit.delete(endpoint)
        else:
            endpoint += "/delete"
            return self.gerrit.post(endpoint, json=input_, headers=self.gerrit.default_headers)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
import inspect
import logging
from functools import wraps


def params_creator(tuples, pattern_types, pattern_dispatcher):
    p, v = None, None
    if pattern_dispatcher is not None and pattern_dispatcher:
        for item in pattern_types:
            if item in pattern_dispatcher:
                p, v = pattern_types[item], pattern_dispatcher[item]
                break
        else:
            k = list(pattern_types.keys())
            raise ValueError("Pattern types can be either " + ", ".join(k[:-1]) + " or " + k[-1])

    params = {
        k: v
        for k, v in tuples + ((p, v),)
        if v is not None
    }

    return params


def check(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        sig = inspect.signature(fn)
        params = sig.parameters
        argspec = list(params.keys())

        # check args and kwargs type
        for i, p in enumerate(args):
            arg_type = fn.__annotations__.get(argspec[i], None)
            if arg_type and not isinstance(p, arg_type):
                raise TypeError(
                    f"{argspec[i]} should be {arg_type.__name__}, not {type(p).__name__}"
                )
        for k, v in kwargs.items():
            kwarg_type = fn.__annotations__.get(k, None)
            if kwarg_type and not isinstance(v, kwarg_type):
                raise TypeError(f"{k} should be {kwarg_type.__name__}, not {type(v).__name__}")

        result = fn(*args, **kwargs)

        # check return type
        return_type = fn.__annotations__.get("return", None)
        if return_type and not isinstance(result, return_type):
            raise TypeError(
                f"{fn.__name__} should return {return_type.__name__}, not {type(result).__name__}"
            )

        return result

    return wrapper


# create logger
logger = logging.getLogger("gerrit")
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logging.root.setLevel(logging.NOTSET)
console = logging.StreamHandler()
console.setLevel(logging.INFO)
console.setFormatter(formatter)
logger.addHandler(console)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi

from gerrit.utils.models import BaseModel


class GerritChangeMessage(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = f"/changes/{self.change}/messages/{self.id}"

    def delete(self, input_=None):
        """
        Deletes a change message.
        Note that only users with the Administrate Server global capability are permitted to
        delete a change message.

        .. code-block:: python

            input_ = {
                "reason": "spam"
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            message = change.messages.get("babf4c5dd53d7a11080696efa78830d0a07762e6")
            result = message.delete(input_)
            # or
            result = message.delete()

        :param input_: the DeleteChangeMessageInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #delete-change-message-input
        :return:
        """
        if input_ is None:
            self.gerrit.delete(self.endpoint)
        else:
            result = self.gerrit.post(self.endpoint + "/delete",
                                      json=input_, headers=self.gerrit.default_headers)
            change = self.gerrit.changes.get(self.change)
            return change.messages.get(result.get("id"))


class GerritChangeMessages(object):
    def __init__(self, change, gerrit):
        self.change = change
        self.gerrit = gerrit
        self.endpoint = f"/changes/{self.change}/messages"

    def list(self):
        """
        Lists all the messages of a change including detailed account information.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        return GerritChangeMessage.parse_list(result, change=self.change, gerrit=self.gerrit)

    def get(self, id_):
        """
        Retrieves a change message including detailed account information.

        :param id_: change message id
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{id_}")
        return GerritChangeMessage(json=result, change=self.change, gerrit=self.gerrit)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi

from gerrit.utils.models import BaseModel


class GerritChangeReviewer(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "username"
        self.endpoint = f"/changes/{self.change}/reviewers/{self.username}"

    def delete(self, input_=None):
        """
        Deletes a reviewer from a change.
        Deleting a reviewer also removes that user from the attention set.

        .. code-block:: python

            input_ = {
                "notify": "NONE"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            reviewer = change.reviewers.get('john.doe')
            reviewer.delete(input_)
            # or
            reviewer.delete()

        :param input_: the DeleteReviewerInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #delete-reviewer-input
        :return:
        """
        if input_ is None:
            self.gerrit.delete(self.endpoint)
        else:
            self.gerrit.post(self.endpoint + "/delete",
                             json=input_, headers=self.gerrit.default_headers)

    def list_votes(self):
        """
        Lists the votes for a specific reviewer of the change.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/votes/")

    def delete_vote(self, label, input_=None):
        """
        Deletes a single vote from a change.
        Note, that even when the last vote of a reviewer is removed the reviewer itself is still
        listed on the change.

        .. code-block:: python

            input_ = {
                "notify": "NONE"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            reviewer = change.reviewers.get('john.doe')
            reviewer.delete_vote('Code-Review', input_)
            # or
            reviewer.delete_vote('Code-Review')

        :param label:
        :param input_: the DeleteVoteInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #delete-vote-input
        :return:
        """
        endpoint = self.endpoint + f"/votes/{label}"
        if input_ is None:
            self.gerrit.delete(endpoint)
        else:
            self.gerrit.post(endpoint + "/delete", json=input_, headers=self.gerrit.default_headers)


class GerritChangeReviewers(object):
    def __init__(self, change, gerrit):
        self.change = change
        self.gerrit = gerrit
        self.endpoint = f"/changes/{self.change}/reviewers"

    def list(self):
        """
        Lists the reviewers of a change.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/")
        return GerritChangeReviewer.parse_list(result, change=self.change, gerrit=self.gerrit)

    def get(self, account):
        """
        Retrieves a reviewer of a change.

        :param account: _account_id, name, username or email
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{account}")
        if result:
            return GerritChangeReviewer(json=result[0], change=self.change, gerrit=self.gerrit)

    def add(self, input_):
        """
        Adds one user or all members of one group as reviewer to the change.

        Users can be moved from reviewer to CC and vice versa. This means if a user is added as
        CC that is already a reviewer on the change, the reviewer state of that user is updated
        to CC. If a user that is already a CC on the change is added as reviewer, the reviewer
        state of that user is updated to reviewer.

        Adding a new reviewer also adds that reviewer to the attention set, unless the change is
        work in progress. Also, moving a reviewer to CC removes that user from the attention set.

        .. code-block:: python

            input_ = {
                "reviewer": "john.doe"
            }

            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            new_reviewer = change.reviewers.add(input_)

        :param input_: the ReviewerInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#reviewer-input
        :return:
        """
        return self.gerrit.post(self.endpoint, json=input_, headers=self.gerrit.default_headers)



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name="python-gerrit-api",
    # https://packaging.python.org/en/latest/single_source_version.html
    version="2.1.2",
    description="Python wrapper for the Gerrit REST API.",
    long_description=long_description,
    url="https://github.com/shijl0925/python-gerrit-api",
    author="Jialiang Shi",
    author_email="kevin09254930sjl@gmail.com",
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development",
    ],
    keywords="api gerrit client wrapper",
    packages=find_packages(exclude=["contrib", "docs", "test*"]),
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=required,
    extras_require={},
    package_data={},
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    data_files=[],
    # Test suite (required for Py2)
    test_suite="tests",
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel


class GerritAccountSSHKey(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "seq"
        self.endpoint = f"/accounts/{self.username}/sshkeys"

    def delete(self):
        """
        Deletes an SSH key of a user.

        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{str(self.seq)}")


class GerritAccountSSHKeys(object):
    def __init__(self, username, gerrit):
        self.username = username
        self.gerrit = gerrit
        self.endpoint = f"/accounts/{self.username}/sshkeys"

    def list(self):
        """
        Returns the SSH keys of an account.

        :return:
        """
        result = self.gerrit.get(self.endpoint)
        return GerritAccountSSHKey.parse_list(result, username=self.username, gerrit=self.gerrit)

    def get(self, seq):
        """
        Retrieves an SSH key of a user.

        :param seq: SSH key id
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{str(seq)}")
        return GerritAccountSSHKey(json=result, username=self.username, gerrit=self.gerrit)

    def add(self, ssh_key):
        """
        Adds an SSH key for a user.
        The SSH public key must be provided as raw content in the request body.

        :param ssh_key: SSH key raw content
        :return:
        """
        result = self.gerrit.post(self.endpoint,
                                  data=ssh_key, headers={"Content-Type": "plain/text"})
        return GerritAccountSSHKey(json=result, username=self.username, gerrit=self.gerrit)

    def delete(self, seq):
        """
        Deletes an SSH key of a user.

        :param seq: SSH key id
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{str(seq)}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus

from gerrit.utils.models import BaseModel


class GerritProjectCommit(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "commit"
        self.endpoint = f"/projects/{self.project}/commits/{self.commit}"

    def get_include_in(self):
        """
        Retrieves the branches and tags in which a change is included.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/in")

    def get_file_content(self, file):
        """
        Gets the content of a file from a certain commit.

        :param file: the file path
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/files/{quote_plus(file)}/content")

    def cherry_pick(self, input_):
        """
        Cherry-picks a commit of a project to a destination branch.

        .. code-block:: python

            input_ = {
                "message": "Implementing Feature X",
                "destination": "release-branch"
            }
            result = commit.cherry_pick(input_)

        :param input_: the CherryPickInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html
          #cherrypick-input
        :return:  the resulting cherry-picked change
        """
        result = self.gerrit.post(
            self.endpoint + "/cherrypick", json=input_, headers=self.gerrit.default_headers)
        return self.gerrit.changes.get(result.get("id"))

    def list_change_files(self):
        """
        Lists the files that were modified, added or deleted in a commit.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/files/")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi
from gerrit.utils.models import BaseModel
from gerrit.accounts.emails import GerritAccountEmails
from gerrit.accounts.ssh_keys import GerritAccountSSHKeys
from gerrit.accounts.gpg_keys import GerritAccountGPGKeys


class GerritAccount(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.entity_name = "username"
        self.endpoint = f"/accounts/{self.username}"

    def get_name(self):
        """
        Retrieves the full name of an account.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/name")

    def set_name(self, input_):
        """
        Sets the full name of an account.
        Some realms may not allow to modify the account name.
        In this case the request is rejected with '405 Method Not Allowed'.

        .. code-block:: python

            input_ = {
                "name": "Keven Shi"
            }

            account = client.accounts.get('kevin.shi')
            result = account.set_name(input_)


        :param input_: the AccountNameInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #account-name-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/name",
                               json=input_, headers=self.gerrit.default_headers)

    def delete_name(self):
        """
        Deletes the name of an account.
        Some realms may not allow to delete the account name.
        In this case the request is rejected with '405 Method Not Allowed'.

        :return:
        """
        self.gerrit.delete(self.endpoint + "/name")

    def get_status(self):
        """
        Retrieves the status of an account.
        If the account does not have a status an empty string is returned.

        :getter: Retrieves the status of an account.
        :setter: Sets the status of an account

        :return:
        """
        return self.gerrit.get(self.endpoint + "/status")

    def set_status(self, status):
        """
        Sets the status of an account.

        :param status: account status
        :return:
        """
        input_ = {"status": status}
        return self.gerrit.put(self.endpoint + "/status",
                               json=input_, headers=self.gerrit.default_headers)

    def set_username(self, input_):
        """
        Sets the username of an account.
        Some realms may not allow to modify the account username.
        In this case the request is rejected with '405 Method Not Allowed'.

        .. code-block:: python

            input_ = {
                "username": "shijl0925.shi"
            }

            account = client.accounts.get('kevin.shi')
            result = account.set_username(input_)

        :param input_: the UsernameInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #username-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/username",
                               json=input_, headers=self.gerrit.default_headers)

    def set_displayname(self, input_):
        """
        Sets the display name of an account.
        support this method since v3.2.0

        .. code-block:: python

            input_ = {
                "display_name": "Kevin"
            }

            account = client.accounts.get('kevin.shi')
            result = account.set_displayname(input_)

        :param input_: the DisplayNameInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #display-name-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/displayname",
                               json=input_, headers=self.gerrit.default_headers)

    def get_active(self):
        """
        Checks if an account is active.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/active")

    def set_active(self):
        """
        Sets the account state to active.

        :return:
        """
        self.gerrit.put(self.endpoint + "/active")

    def delete_active(self):
        """
        Sets the account state to inactive.
        If the account was already inactive the response is '409 Conflict'.

        :return:
        """
        self.gerrit.delete(self.endpoint + "/active")

    def set_http_password(self, input_):
        """
        Sets/Generates the HTTP password of an account.

        .. code-block:: python

            input_ = {
                "generate": 'true',
                "http_password": "the_password"
            }

            account = client.accounts.get('kevin.shi')
            result = account.set_http_password(input_)

        :param input_: the HttpPasswordInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #http-password-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/password.http",
                               json=input_, headers=self.gerrit.default_headers)

    def delete_http_password(self):
        """
        Deletes the HTTP password of an account.

        :return:
        """
        self.gerrit.delete(self.endpoint + "/password.http")

    def get_oauth_token(self):
        """
        Returns a previously obtained OAuth access token.
        If there is no token available, or the token has already expired, '404 Not Found' is
        returned as response.
        Requests to obtain an access token of another user are rejected with '403 Forbidden'.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/oauthtoken")

    @property
    def emails(self):
        return GerritAccountEmails(username=self.username, gerrit=self.gerrit)

    @property
    def ssh_keys(self):
        return GerritAccountSSHKeys(username=self.username, gerrit=self.gerrit)

    @property
    def gpg_keys(self):
        return GerritAccountGPGKeys(username=self.username, gerrit=self.gerrit)

    def list_capabilities(self):
        """
        Returns the global capabilities that are enabled for the specified user.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/capabilities")

    def check_capability(self, capability):
        """
        Checks if a user has a certain global capability.

        :param capability:
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/capabilities/{capability}")

    @property
    def groups(self):
        """
        Lists all groups that contain the specified user as a member.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/groups")
        return [self.gerrit.groups.get(item.get("id")) for item in result]

    def get_avatar(self):
        """
        Retrieves the avatar image of the user, requires avatars-gravatar plugin.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/avatar")

    def get_avatar_change_url(self):
        """
        Retrieves the avatar image of the user, requires avatars-gravatar plugin.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/avatar.change.url")

    def get_user_preferences(self):
        """
        Retrieves the user’s preferences.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/preferences")

    def set_user_preferences(self, input_):
        """
        Sets the user’s preferences.

        .. code-block:: python

            input_ = {
                "changes_per_page": 50,
                "show_site_header": true,
                "use_flash_clipboard": true,
                "expand_inline_diffs": true,
                "download_command": "CHECKOUT",
                "date_format": "STD",
                "time_format": "HHMM_12",
                "size_bar_in_change_table": true,
                "review_category_strategy": "NAME",
                "diff_view": "SIDE_BY_SIDE",
                "mute_common_path_prefixes": true,
            }

            account = client.accounts.get('kevin.shi')
            result = account.set_user_preferences(input_)

        :param input_: the PreferencesInput entity，
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #preferences-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/preferences",
                               json=input_, headers=self.gerrit.default_headers)

    def get_diff_preferences(self):
        """
        Retrieves the diff preferences of a user.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/preferences.diff")

    def set_diff_preferences(self, input_):
        """
        Sets the diff preferences of a user.

        .. code-block:: python

            input_ = {
                "context": 10,
                "theme": "ECLIPSE",
                "ignore_whitespace": "IGNORE_ALL",
                "intraline_difference": true,
                "line_length": 100,
                "cursor_blink_rate": 500,
                "show_line_endings": true,
                "show_tabs": true,
                "show_whitespace_errors": true,
                "syntax_highlighting": true,
                "tab_size": 8,
                "font_size": 12
            }

            account = client.accounts.get('kevin.shi')
            result = account.set_diff_preferences(input_)

        :param input_: the DiffPreferencesInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #diff-preferences-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/preferences.diff",
                               json=input_, headers=self.gerrit.default_headers)

    def get_edit_preferences(self):
        """
        Retrieves the edit preferences of a user.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/preferences.edit")

    def set_edit_preferences(self, input_):
        """
        Sets the edit preferences of a user.

        .. code-block:: python

            input_ = {
                "theme": "ECLIPSE",
                "key_map_type": "VIM",
                "tab_size": 4,
                "line_length": 80,
                "indent_unit": 2,
                "cursor_blink_rate": 530,
                "hide_top_menu": true,
                "show_tabs": true,
                "show_whitespace_errors": true,
                "syntax_highlighting": true,
                "hide_line_numbers": true,
                "match_brackets": true,
                "line_wrapping": false,
                "auto_close_brackets": true
            }

            account = client.accounts.get('kevin.shi')
            result = account.set_edit_preferences(input_)

        :param input_: the EditPreferencesInfo entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #edit-preferences-info
        :return:
        """
        return self.gerrit.put(self.endpoint + "/preferences.edit",
                               json=input_, headers=self.gerrit.default_headers)

    def get_watched_projects(self):
        """
        Retrieves all projects a user is watching.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/watched.projects")

    def modify_watched_projects(self, input_):
        """
        Add new projects to watch or update existing watched projects.
        Projects that are already watched by a user will be updated with the provided configuration.

        .. code-block:: python

            input_ = [
                {
                    "project": "Test Project 1",
                    "notify_new_changes": true,
                    "notify_new_patch_sets": true,
                    "notify_all_comments": true,
                }
            ]

            account = client.accounts.get('kevin.shi')
            result = account.modify_watched_projects(input_)

        :param input_: the ProjectWatchInfo entities as list
        :return:
        """
        return self.gerrit.post(self.endpoint + "/watched.projects",
                                json=input_, headers=self.gerrit.default_headers)

    def delete_watched_projects(self, input_):
        """
        Projects posted to this endpoint will no longer be watched.

        .. code-block:: python

            input_ = [
                {
                    "project": "Test Project 1",
                    "filter": "branch:master"
                }
            ]

            account = client.accounts.get('kevin.shi')
            result = account.delete_watched_projects(input_)

        :param input_: the watched projects as list
        :return:
        """
        self.gerrit.post(self.endpoint + "/watched.projects:delete",
                         json=input_, headers=self.gerrit.default_headers)

    def get_external_ids(self):
        """
        Retrieves the external ids of a user account.
        Only external ids belonging to the caller may be requested.
        Users that have Modify Account can request external
        ids that belong to other accounts.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/external.ids")

    def delete_external_ids(self, input_):
        """
        Delete a list of external ids for a user account.
        Only external ids belonging to the caller may be deleted. Users that have Modify Account
        can delete external ids that belong to other accounts.

        .. code-block:: python

            input_ = [
                "mailto:john.doe@example.com"
            ]

            account = client.accounts.get('kevin.shi')
            result = account.delete_external_ids(input_)

        :param input_: the external ids as list
        :return:
        """
        self.gerrit.post(self.endpoint + "/external.ids:delete",
                         json=input_, headers=self.gerrit.default_headers)

    def list_contributor_agreements(self):
        """
        Gets a list of the user’s signed contributor agreements.

        :return:
        """
        return self.gerrit.get(self.endpoint + "/agreements")

    def sign_contributor_agreement(self, input_):
        """
        Signs a contributor agreement.

        .. code-block:: python

            input_ = {
                "name": "Individual"
            }
            account = client.accounts.get('kevin.shi')
            result = account.sign_contributor_agreement(input_)

        :param input_: the ContributorAgreementInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #contributor-agreement-input
        :return:
        """
        return self.gerrit.put(self.endpoint + "/agreements",
                               json=input_, headers=self.gerrit.default_headers)


    def delete_draft_comments(self, input_):
        """
        Deletes some or all of a user’s draft comments.

        .. code-block:: python

            input_ = {
                "query": "is:abandoned"
            }
            account = client.accounts.get('kevin.shi')
            result = account.delete_draft_comments(input_)

        :param input_: the DeleteDraftCommentsInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html
          #delete-draft-comments-input
        :return:
        """
        return self.gerrit.post(self.endpoint + "/drafts:delete",
                                json=input_, headers=self.gerrit.default_headers)

    def index(self):
        """
        Adds or updates the account in the secondary index.

        :return:
        """
        self.gerrit.post(self.endpoint + "/index")

    def get_default_starred_changes(self):
        """
        Gets the changes that were starred with the default star by the identified user account.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/starred.changes")
        return [self.gerrit.changes.get(item.get("id")) for item in result]

    def put_default_star_on_change(self, id_):
        """
        Star a change with the default label.

        :param id_: change id
        :return:
        """
        self.gerrit.put(self.endpoint + f"/starred.changes/{id_}")

    def remove_default_star_from_change(self, id_):
        """
        Remove the default star label from a change. This stops notifications.

        :param id_: change id
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/starred.changes/{id_}")

    def get_starred_changes(self):
        """
        Gets the changes that were starred with any label by the identified user account.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/stars.changes")
        return [self.gerrit.changes.get(item.get("id")) for item in result]

    def get_star_labels_from_change(self, id_):
        """
        Get star labels from a change.

        :param id_: change id
        :return:
        """
        return self.gerrit.get(self.endpoint + f"/stars.changes/{id_}")

    def update_star_labels_on_change(self, id_, input_):
        """
        Update star labels on a change.

        .. code-block:: python

            input_ = {
                "add": ["blue", "red"],
                "remove": ["yellow"]
            }

            account = client.accounts.get('kevin.shi')
            change_id = "myProject~master~I8473b95934b5732ac55d26311a706c9c2bde9940"
            result = account.update_star_labels_on_change(change_id, input_)


        :param id_: change id
        :param input_: the StarsInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-accounts.html#stars-input
        :return:
        """
        return self.gerrit.post(self.endpoint + f"/stars.changes/{id_}",
                                json=input_, headers=self.gerrit.default_headers)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi

from gerrit.utils.models import BaseModel


class GerritChangeRevisionDraft(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = f"/changes/{self.change}/revisions/{self.revision}/drafts/{self.id}"

    def update(self, input_):
        """
        Updates a draft comment on a revision.

        .. code-block:: python

            input_ = {
                "path": "sonarqube/cloud/duplications.py",
                "line": 25,
                "message": "[nit] trailing whitespace"
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision('3848807f587dbd3a7e61723bbfbf1ad13ad5a00a')
            draft = revision.drafts.get('89f04e8c_9b7fd51d')
            result = draft.update(input_)

        :param input_: the CommentInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#comment-input
        :return:
        """
        result = self.gerrit.put(self.endpoint, json=input_, headers=self.gerrit.default_headers)
        return GerritChangeRevisionDraft.parse(
            result, change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    def delete(self):
        """
        Deletes a draft comment from a revision.

        :return:
        """
        self.gerrit.delete(self.endpoint)


class GerritChangeRevisionDrafts(object):
    def __init__(self, change, revision, gerrit):
        self.change = change
        self.revision = revision
        self.gerrit = gerrit
        self.endpoint = self.gerrit.get(f"/changes/{self.change}/revisions/{self.revision}/drafts")

    def list(self):
        """
        Lists the draft comments of a revision that belong to the calling user.

        :return:
        """
        result = self.gerrit.get(f"/changes/{self.change}/revisions/{self.revision}/drafts")
        drafts = []
        for key, value in result.items():
            for item in value:
                draft = item
                draft.update({"path": key})
                drafts.append(draft)
        return GerritChangeRevisionDraft.parse_list(
            drafts, change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    def get(self, id_):
        """
        Retrieves a draft comment of a revision that belongs to the calling user.

        :param id_: the draft comment id
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{id_}")
        return GerritChangeRevisionDraft.parse(
            result, change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    def create(self, input_):
        """
        Creates a draft comment on a revision.

        .. code-block:: python

            input_ = {
                "path": "sonarqube/cloud/duplications.py",
                "line": 15,
                "message": "[nit] trailing whitespace"
            }
            change = client.changes.get('Project~stable~I10394472cbd17dd12454f229e4f6de00b143a444')
            revision = change.get_revision('3848807f587dbd3a7e61723bbfbf1ad13ad5a00a')
            new_draft = revision.drafts.create(input_)

        :param input_: the CommentInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-changes.html#comment-input
        :return:
        """
        result = self.gerrit.put(self.endpoint, json=input_, headers=self.gerrit.default_headers)
        return GerritChangeRevisionDraft.parse(
            result, change=self.change, revision=self.revision, gerrit=self.gerrit
        )

    def delete(self, id_):
        """
        Deletes a draft comment from a revision.

        :param id_: the draft comment id
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{id_}")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @Author: Jialiang Shi

from gerrit.utils.models import BaseModel


class GerritProjectDashboard(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def delete(self):
        """
        Deletes a project dashboard.

        :return:
        """
        self.gerrit.delete(f"/projects/{self.project}/dashboards/{self.id}")


class GerritProjectDashboards(object):
    def __init__(self, project, gerrit):
        self.project = project
        self.gerrit = gerrit
        self.endpoint = f"/projects/{self.project}/dashboards"

    def list(self):
        """
        List custom dashboards for a project.

        :return:
        """
        result = self.gerrit.get(self.endpoint + "/")
        return GerritProjectDashboard.parse_list(result, project=self.project, gerrit=self.gerrit)

    def create(self, id_, input_):
        """
        Creates a project dashboard, if a project dashboard with the given dashboard ID doesn't
        exist yet.

        .. code-block:: python

            input_ = {
                "id": "master:closed",
                "commit_message": "Define the default dashboard"
            }
            new_dashboard = project.dashboards.create('master:closed', input_)


        :param id_: the dashboard id
        :param input_: the DashboardInput entity,
          https://gerrit-review.googlesource.com/Documentation/rest-api-projects.html
          #dashboard-input
        :return:
        """
        result = self.gerrit.put(
            self.endpoint + f"/{id_}", json=input_, headers=self.gerrit.default_headers)
        return GerritProjectDashboard(json=result, project=self.project, gerrit=self.gerrit)

    def get(self, id_):
        """
        Retrieves a project dashboard. The dashboard can be defined on that project or be inherited
        from a parent project.

        :param id_: the dashboard id
        :return:
        """
        result = self.gerrit.get(self.endpoint + f"/{id_}")
        return GerritProjectDashboard(json=result, project=self.project, gerrit=self.gerrit)

    def delete(self, id_):
        """
        Deletes a project dashboard.

        :param id_: the dashboard id
        :return:
        """
        self.gerrit.delete(self.endpoint + f"/{id_}")
